"""Define key components of main window"""

from enum import Enum, auto
from idlelib.redirector import WidgetRedirector  # type: ignore[import-not-found]
import logging
import os.path
import time
import tkinter as tk
from tkinter import ttk, messagebox
from typing import Any, Callable, Optional

from PIL import Image, ImageTk, ImageChops

from guiguts.maintext import MainText, maintext
from guiguts.preferences import preferences, PrefKey, PersistentBoolean
from guiguts.root import Root, root
from guiguts.utilities import (
    is_mac,
    is_x11,
    bell_set_callback,
    process_accel,
    process_label,
    IndexRowCol,
    sound_bell,
)
from guiguts.widgets import (
    ToplevelDialog,
    mouse_bind,
    ToolTip,
    themed_style,
    theme_set_tk_widget_colors,
    Busy,
)

logger = logging.getLogger(__package__)

TEXTIMAGE_WINDOW_ROW = 0
TEXTIMAGE_WINDOW_COL = 0
SEPARATOR_ROW = 1
SEPARATOR_COL = 0
STATUS_ROW = 2
STATUS_COL = 0
MIN_PANE_WIDTH = 20


class Menu(tk.Menu):
    """Extend ``tk.Menu`` to make adding buttons with accelerators simpler."""

    def __init__(self, parent: tk.Widget, label: str, **kwargs: Any) -> None:
        """Initialize menu and add to parent

        Args:
            parent: Parent menu/menubar, or another widget if context menu.
            label: Label string for menu, including tilde for keyboard
              navigation, e.g. "~File".
            **kwargs: Optional additional keywords args for ``tk.Menu``.
        """

        super().__init__(parent, **kwargs)
        command_args: dict[str, Any] = {"menu": self}
        if label:
            (label_tilde, label_txt) = process_label(label)
            command_args["label"] = label_txt
            if label_tilde >= 0:
                command_args["underline"] = label_tilde
        # Only needs cascade if a child of menu/menubar, not if a context popup menu
        if isinstance(parent, tk.Menu):
            parent.add_cascade(command_args)

    def add_button(
        self,
        label: str,
        handler: str | Callable[[], Any],
        accel: str = "",
        force_main_only: bool = False,
    ) -> None:
        """Add a button to the menu.

        Args:
            label: Label string for button, including tilde for keyboard
              navigation, e.g. "~Save".
            handler: Callback function or built-in virtual event,
              e.g. "<<Cut>>", in which case button will generate that event.
            accel: String describing optional accelerator key, used when a
              callback function is passed in as ``handler``. Will be displayed
              on the button, and will be bound to the same action as the menu
              button. "Cmd/Ctrl" means `Cmd` key on Mac; `Ctrl` key on
              Windows/Linux.
            force_main_only: True to only bind to the main window, not other dialogs
        """
        (label_tilde, label_txt) = process_label(label)
        (accel, key_event) = process_accel(accel)
        if isinstance(handler, str):
            bind_all = False

            # Handler is built-in virtual event, which needs to be
            # generated by button click
            def command() -> None:
                widget = root().focus_get()
                if widget is not None:
                    widget.event_generate(handler)

        else:
            bind_all = True
            # Handler is function, so may need key binding
            command = handler

        if force_main_only:
            bind_all = False

        # If key binding given, then bind it
        if accel:
            # Note that when run via a key binding, Tk will pass in an event arg
            # to the callback function, but this arg is never used, and is absorbed
            # by the lambda function below
            # Key is bound to all widgets, except for built-in virtual events.
            maintext().key_bind(key_event, lambda _event: command(), bind_all=bind_all)

        command_args = {
            "label": label_txt,
            "command": command,
            "accelerator": accel,
        }
        if label_tilde >= 0:
            command_args["underline"] = label_tilde
        self.add_command(command_args)

    def add_checkbox(
        self,
        label: str,
        bool_var: tk.BooleanVar,
        handler_on: Optional[Callable[[], None]] = None,
        handler_off: Optional[Callable[[], None]] = None,
        accel: str = "",
    ) -> None:
        """Add a button to the menu.

        Args:
            label: Label string for button, including tilde for keyboard
              navigation, e.g. "~Save".
            handler_on: Callback function for when checkbox gets checked
            handler_off: Callback function for when checkbox gets checked
            bool_var: Tk variable to keep track of state or set it from elsewhere
            accel: String describing optional accelerator key, used when a
              callback function is passed in as ``handler``. Will be displayed
              on the button, and will be bound to the same action as the menu
              button. "Cmd/Ctrl" means `Cmd` key on Mac; `Ctrl` key on
              Windows/Linux.
        """
        (label_tilde, label_txt) = process_label(label)
        (accel, key_event) = process_accel(accel)

        # If key binding given, then bind it
        if accel:

            def accel_command(_event: tk.Event) -> None:
                """Command to simulate checkbox click via shortcut key.

                Because key hasn't been clicked, variable hasn't been toggled.
                """
                bool_var.set(not bool_var.get())
                checkbox_clicked()

            maintext().key_bind(key_event, accel_command, bind_all=True)

        def checkbox_clicked() -> None:
            """Callback when checkbox is clicked.

            Call appropriate handler depending on setting."""
            if bool_var.get():
                if handler_on is not None:
                    handler_on()
            else:
                if handler_off is not None:
                    handler_off()

        command_args = {
            "label": label_txt,
            "command": checkbox_clicked,
            "variable": bool_var,
            "accelerator": accel,
        }
        if label_tilde >= 0:
            command_args["underline"] = label_tilde
        self.add_checkbutton(command_args)

    def add_cut_copy_paste(self, read_only: bool = False) -> None:
        """Add cut/copy/paste buttons to this menu"""
        if not read_only:
            self.add_button("Cu~t", "<<Cut>>", "Cmd/Ctrl+X")
        self.add_button("~Copy", "<<Copy>>", "Cmd/Ctrl+C")
        if not read_only:
            self.add_button("~Paste", "<<Paste>>", "Cmd/Ctrl+V")
        self.add_separator()
        self.add_button("Select ~All", "<<SelectAll>>", "Cmd/Ctrl+A")


class AutoImageState(Enum):
    """Enum class to store AutoImage states to facilitate pausing
    and restarting AutoImage when needed."""

    NORMAL = auto()
    PAUSED = auto()
    RESTARTING = auto()


class MainImage(tk.Frame):
    """MainImage is a Frame, containing a Canvas which can display a png/jpeg file.

    Also contains scrollbars, and can be scrolled with mousewheel (vertically),
    Shift-mousewheel (horizontally) and zoomed with Control-mousewheel.

    MainImage can be docked or floating. Floating is not supported with ttk.Frame,
    hence inherits from tk.Frame.

    Attributes:
        hbar: Horizontal scrollbar.
        vbar: Vertical scrollbar.
        canvas: Canvas widget.
        image: Whole loaded image (or None)
        image_scale: Zoom scale at which image should be drawn.
        scale_delta: Ratio to multiply/divide scale when Control-scrolling mouse wheel.
    """

    def __init__(
        self,
        parent: tk.PanedWindow,
        hide_func: Callable[[], None],
        float_func: Callable[[Optional[tk.Event]], None],
        dock_func: Callable[[Optional[tk.Event]], None],
    ) -> None:
        """Initialize the MainImage to contain an empty Canvas with scrollbars.

        Args:
            hide_func: Function to hide the image viewer.
            float_func: Function to float the image viewer.
            dock_func: Function to dock the image viewer.
        """
        tk.Frame.__init__(self, parent, name=" Image Viewer")
        self.parent = parent
        self.hide_func = hide_func
        self.float_func = float_func
        self.dock_func = dock_func
        self.allow_geometry_storage = False

        # Introduce an apparently superfluous Frame to contain everything.
        # This is because when MainImage is undocked, Tk converts it to a
        # Toplevel widget. Any bindings on the MainImage are then reapplied
        # to the Toplevel widget, and Toplevel widgets apply their bindings
        # to all their children. So an Enter/Leave binding on MainImage would
        # cause an Enter/Leave to trigger when the mouse enters/leaves a button,
        # for example. Binding to top_frame instead means that the binding
        # doesn't get reapplied to the Toplevel widget, and it continues to
        # behave as required.
        top_frame = ttk.Frame(self)
        top_frame.grid(row=0, column=0, columnspan=2, sticky="NSEW")
        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)
        control_frame = ttk.Frame(top_frame)
        control_frame.grid(row=0, column=0, columnspan=2, sticky="NSEW")
        control_frame.columnconfigure(8, weight=1)

        min_button_width = 1 if is_mac() else 2

        self.prev_img_button = ttk.Button(
            control_frame,
            text="<",
            width=min_button_width,
            takefocus=False,
            command=lambda: self.next_image(reverse=True),
        )
        self.prev_img_button.grid(row=0, column=0, sticky="NSEW")

        self.next_img_button = ttk.Button(
            control_frame,
            text=">",
            width=min_button_width,
            takefocus=False,
            command=self.next_image,
        )
        self.next_img_button.grid(row=0, column=1, sticky="NSEW")

        self.zoom_in_btn = ttk.Button(
            control_frame,
            text="+",
            width=min_button_width,
            takefocus=False,
            command=lambda: self.image_zoom(zoom_in=True),
        )
        self.zoom_in_btn.grid(row=0, column=2, sticky="NSEW", padx=(5, 0))

        self.zoom_out_btn = ttk.Button(
            control_frame,
            text="-",
            width=min_button_width,
            takefocus=False,
            command=lambda: self.image_zoom(zoom_in=False),
        )
        self.zoom_out_btn.grid(row=0, column=3, sticky="NSEW")

        self.ftw_btn = ttk.Button(
            control_frame,
            text="← Fit →",
            takefocus=False,
            command=self.image_zoom_to_width,
        )
        self.ftw_btn.grid(row=0, column=4, sticky="NSEW", padx=(5, 0))

        self.fth_btn = ttk.Button(
            control_frame,
            text="↑ Fit ↓",
            takefocus=False,
            command=self.image_zoom_to_height,
        )
        self.fth_btn.grid(row=0, column=5, sticky="NSEW")

        self.invert_btn = ttk.Checkbutton(
            control_frame,
            text="Invert",
            takefocus=False,
            command=self.show_image,
            variable=PersistentBoolean(PrefKey.IMAGE_INVERT),
        )
        self.invert_btn.grid(row=0, column=6, sticky="NSEW", padx=(5, 0))

        self.dock_btn = ttk.Checkbutton(
            control_frame,
            text="Dock",
            takefocus=False,
            command=self.set_image_docking,
            variable=root().image_window_docked_state,
        )
        self.dock_btn.grid(row=0, column=7, sticky="NSEW", padx=(5, 0))

        self.close_btn = ttk.Button(
            control_frame,
            text="×",
            width=min_button_width,
            takefocus=False,
            command=self.hide_func,
        )
        self.close_btn.grid(row=0, column=8, sticky="NSE")

        # Separate bindings needed for docked (root) and floated (self) states
        for widget in (root(), self):
            _, cp = process_accel("Cmd/Ctrl+plus")
            _, ce = process_accel("Cmd/Ctrl+equal")
            _, cm = process_accel("Cmd/Ctrl+minus")
            _, c0 = process_accel("Cmd/Ctrl+0")
            widget.bind(cp, lambda _: self.zoom_in_btn.invoke())
            widget.bind(ce, lambda _: self.zoom_in_btn.invoke())
            widget.bind(cm, lambda _: self.zoom_out_btn.invoke())
            widget.bind(c0, lambda _: self.fth_btn.invoke())

        self.hbar = ttk.Scrollbar(top_frame, orient=tk.HORIZONTAL)
        self.hbar.grid(row=3, column=0, sticky="EW")
        self.hbar.configure(command=self.scroll_x)
        self.vbar = ttk.Scrollbar(top_frame, orient=tk.VERTICAL)
        self.vbar.grid(row=2, column=1, sticky="NS")
        self.vbar.configure(command=self.scroll_y)

        self.canvas = tk.Canvas(
            top_frame,
            xscrollcommand=self.hbar.set,
            yscrollcommand=self.vbar.set,
            highlightthickness=0,
            highlightbackground="darkorange",
        )
        self.canvas.grid(row=2, column=0, sticky="NSEW")
        top_frame.rowconfigure(2, weight=1)
        top_frame.columnconfigure(0, weight=1)

        self.canvas.bind("<Configure>", self.handle_configure)
        top_frame.bind("<Enter>", self.handle_enter)
        top_frame.bind("<Leave>", self.handle_leave)
        self.canvas.bind("<ButtonPress-1>", self.move_from)
        self.canvas.bind("<B1-Motion>", self.move_to)
        if is_x11():
            self.canvas.bind("<Control-Button-5>", self.wheel_zoom)
            self.canvas.bind("<Control-Button-4>", self.wheel_zoom)
            self.canvas.bind("<Button-5>", self.wheel_scroll)
            self.canvas.bind("<Button-4>", self.wheel_scroll)
        else:
            _, cm = process_accel("Cmd/Ctrl+MouseWheel")
            self.canvas.bind(cm, self.wheel_zoom)
            self.canvas.bind("<MouseWheel>", self.wheel_scroll)

        self.image_scale = float(preferences.get(PrefKey.IMAGE_SCALE_FACTOR))
        self.scale_delta = 1.1
        self.image: Optional[Image.Image] = None
        self.imageid = None
        self.imagetk = None
        self.filename = ""
        self.width = 0
        self.height = 0
        # May want to pause auto image if user clicks prev/next file buttons
        self._auto_image_state = AutoImageState.NORMAL

    def scroll_y(self, *args: Any, **kwargs: Any) -> None:
        """Scroll canvas vertically and redraw the image"""
        self.canvas.yview(*args, **kwargs)
        self.show_image()

    def scroll_x(self, *args: Any, **kwargs: Any) -> None:
        """Scroll canvas horizontally and redraw the image."""
        self.canvas.xview(*args, **kwargs)
        self.show_image()

    def move_from(self, event: tk.Event) -> None:
        """Remember previous coordinates for dragging with the mouse."""
        self.canvas.scan_mark(event.x, event.y)

    def move_to(self, event: tk.Event) -> None:
        """Drag canvas to the new position."""
        self.canvas.scan_dragto(event.x, event.y, gain=1)
        self.show_image()

    def wheel_zoom(self, event: tk.Event) -> None:
        """Zoom with mouse wheel.

        Args:
            event: Event containing mouse wheel info.
        """
        # Respond to Linux (event.num) or Windows/MacOS (event.delta) wheel event
        if event.num == 5 or event.delta < 0:
            self.zoom_out_btn.invoke()
        if event.num == 4 or event.delta > 0:
            self.zoom_in_btn.invoke()

    def image_zoom(self, zoom_in: bool) -> None:
        """Zoom the image in or out.

        Args:
            zoom_in: True to zoom in, False to zoom out.
        """
        if zoom_in:
            if self.image_scale < 3:
                self.image_scale *= self.scale_delta
        else:
            if self.image_scale > 0.1:
                self.image_scale /= self.scale_delta
        preferences.set(PrefKey.IMAGE_SCALE_FACTOR, self.image_scale)
        self.show_image()

    def image_zoom_to_width(self) -> None:
        """Zoom image to fit to width of image window."""
        if self.imageid is None:
            return
        bbox_image = self.canvas.bbox(self.imageid)
        scale_factor = (
            self.canvas.canvasx(self.canvas.winfo_width()) - self.canvas.canvasx(0)
        ) / (bbox_image[2] - bbox_image[0])
        self.image_zoom_by_factor(scale_factor)

    def image_zoom_to_height(self) -> None:
        """Zoom image to fit to height of image window."""
        if self.imageid is None:
            return
        bbox_image = self.canvas.bbox(self.imageid)
        scale_factor = (
            self.canvas.canvasy(self.canvas.winfo_height()) - self.canvas.canvasy(0)
        ) / (bbox_image[3] - bbox_image[1])
        self.image_zoom_by_factor(scale_factor)

    def image_zoom_by_factor(self, scale_factor: float) -> None:
        """Zoom image by the given scale factor.

        Args:
            scale_factor: Factor to zoom by.
        """
        self.image_scale *= scale_factor
        preferences.set(PrefKey.IMAGE_SCALE_FACTOR, self.image_scale)
        self.canvas.xview_moveto(0.0)
        self.canvas.yview_moveto(0.0)
        self.show_image()

    def show_image(self, _event=None):  # type: ignore[no-untyped-def]
        """Show image on the Canvas"""
        # get image area & remove 1 pixel shift
        if self.image is None:
            return
        self.canvas["background"] = themed_style().lookup("TButton", "background")
        image = self.image
        scaled_width = int(self.image_scale * image.width)
        scaled_height = int(self.image_scale * image.height)
        if preferences.get(PrefKey.IMAGE_INVERT):
            image = ImageChops.invert(self.image)
        if self.imagetk:
            del self.imagetk
        image = image.resize(
            size=(scaled_width, scaled_height), resample=Image.Resampling.LANCZOS
        )
        alpha_value = 180 if maintext().is_dark_theme() else 200
        image.putalpha(alpha_value)  # Reduce contrast by making slightly transparent
        self.imagetk = ImageTk.PhotoImage(image)
        if self.imageid:
            self.canvas.delete(self.imageid)
        self.imageid = self.canvas.create_image(
            0,
            0,
            anchor="nw",
            image=self.imagetk,
        )
        self.canvas.configure(scrollregion=self.canvas.bbox(self.imageid))

    def wheel_scroll(self, evt: tk.Event) -> None:
        """Scroll image up/down using mouse wheel"""
        if evt.state == 0:
            if is_mac():
                self.canvas.yview_scroll(-1 * (evt.delta), "units")
            else:
                self.canvas.yview_scroll(int(-1 * (evt.delta / 120)), "units")
        if evt.state == 1:
            if is_mac():
                self.canvas.xview_scroll(-1 * (evt.delta), "units")
            else:
                self.canvas.xview_scroll(int(-1 * (evt.delta / 120)), "units")
        self.show_image()

    def load_image(self, filename: Optional[str] = None) -> bool:
        """Load or clear the given image file.

        Args:
            filename: Optional name of image file. If none given, clear image.

        Returns:
            True if new file was loaded, False otherwise
        """
        if filename == self.filename:
            return False

        if filename and os.path.isfile(filename):
            self.filename = filename
            self.image = Image.open(filename).convert("RGB")
            self.width, self.height = self.image.size
            self.canvas.yview_moveto(0)
            self.canvas.xview_moveto(0)
            self.show_image()
        else:
            self.clear_image()
        return True

    def clear_image(self) -> None:
        """Clear the image and reset variables accordingly."""
        self.filename = ""
        self.image = None
        if self.imageid:
            self.canvas.delete(self.imageid)
        self.imageid = None

    def set_image_docking(self) -> None:
        """Float/dock image depending on flag."""
        if preferences.get(PrefKey.IMAGE_WINDOW_DOCKED):
            self.dock_func(None)
        else:
            self.float_func(None)

    def is_image_loaded(self) -> bool:
        """Return if an image is currently loaded."""
        return self.image is not None

    def enable_geometry_storage(self) -> None:
        """Permit storage of image viewer geometry.

        Don't want to allow it right at start in case window manager places
        dialog, causing a config event and overwriting the stored geometry.
        """
        try:
            tk.Wm.geometry(self, preferences.get(PrefKey.IMAGE_FLOAT_GEOMETRY))  # type: ignore[call-overload]
        except tk.TclError:
            pass
        self.allow_geometry_storage = True

    def handle_configure(self, _e: tk.Event) -> None:
        """Handle configure event."""
        if preferences.get(PrefKey.IMAGE_WINDOW_DOCKED):
            try:  # In case unlucky timing means it tries to configure during undocking & finds sash doesn't exist
                preferences.set(
                    PrefKey.IMAGE_DOCK_SASH_COORD, self.parent.sash_coord(0)[0]
                )
            except tk.TclError:
                pass
        elif self.allow_geometry_storage:
            try:  # In case unlucky timing means it tries to configure during docking & widget isn't a toplevel
                preferences.set(PrefKey.IMAGE_FLOAT_GEOMETRY, tk.Wm.geometry(self))  # type: ignore[call-overload]
            except tk.TclError:
                pass

    def handle_enter(self, event: tk.Event) -> None:
        """Handle enter event."""
        self.handle_configure(event)
        self.auto_image_state(AutoImageState.NORMAL)

    def handle_leave(self, event: tk.Event) -> None:
        """Handle leave event."""
        self.handle_configure(event)
        # Restart auto img
        if self.auto_image_state() == AutoImageState.PAUSED:
            self.auto_image_state(AutoImageState.RESTARTING)

    def alert_user(self) -> None:
        """Flash the image border if preference enabled."""
        if not preferences.get(PrefKey.IMAGE_VIEWER_ALERT):
            return

        def set_canvas_highlight(thickness: int) -> None:
            """Set highlightthickness for canvas"""
            self.canvas["highlightthickness"] = thickness

        set_canvas_highlight(4)
        self.after(500, lambda: set_canvas_highlight(0))

    def auto_image_state(
        self, value: Optional[AutoImageState] = None
    ) -> AutoImageState:
        """Set or query whether auto_image is paused or restarting."""
        if value is not None:
            self._auto_image_state = value
        return self._auto_image_state

    def next_image(self, reverse: bool = False) -> None:
        """Load the next image alphabetically.

        Args:
            reverse: True to load previous image instead.
        """
        if not self.filename:
            sound_bell()
            return

        # Check current directory is valid
        current_dir = os.path.dirname(self.filename)
        if not os.path.isdir(current_dir):
            logger.error(f"Image directory invalid: {current_dir}")
            return

        current_basename = os.path.basename(self.filename)
        found = False
        for fn in sorted(os.listdir(current_dir), reverse=reverse):
            # Skip non-image files by checking extension
            if os.path.splitext(fn)[1] not in (".jpg", ".gif", ".png"):
                continue
            # If found on previous time through loop, this is the file we want
            if found:
                self.load_image(os.path.join(current_dir, fn))
                self.auto_image_state(AutoImageState.PAUSED)
                return
            if fn == current_basename:
                found = True

        # Reached end of dir listing without finding next file
        sound_bell()


class StatusBar(ttk.Frame):
    """Statusbar at the bottom of the screen.

    Fields in statusbar can be automatically or manually updated.
    """

    def __init__(self, parent: ttk.Frame) -> None:
        """Initialize statusbar within given frame.

        Args:
            parent: Frame to contain status bar.
        """
        super().__init__(parent)
        self.fields: dict[str, ttk.Button] = {}
        self.callbacks: dict[str, Optional[Callable[[], str]]] = {}
        self._update()

    def add(
        self,
        key: str,
        tooltip: str = "",
        update: Optional[Callable[[], str]] = None,
        **kwargs: Any,
    ) -> None:
        """Add field to status bar

        Args:
            key: Key to use to refer to field.
            update: Optional callback function that returns a string.
              If supplied, field will be regularly updated automatically with
              the string returned by ``update()``. If argument not given,
              application is responsible for updating, using ``set(key)``.
        """
        self.fields[key] = ttk.Button(self, takefocus=0, **kwargs)
        self.callbacks[key] = update
        self.fields[key].grid(column=len(self.fields), row=0)
        if tooltip:
            ToolTip(self.fields[key], msg=tooltip)

    def set(self, key: str, value: str) -> None:
        """Set field in statusbar to given value.

        Args:
            key: Key to refer to field.
            value: String to use to update field.
        """
        self.fields[key].config(text=value)

    def _update(self) -> None:
        """Update fields in statusbar that have callbacks. Updates every
        200 milliseconds.
        """
        for key in self.fields:
            func = self.callbacks[key]
            if func is not None:
                self.set(key, func())
        self.after(200, self._update)

    def add_binding(self, key: str, event: str, callback: Callable[[], None]) -> None:
        """Add an action to be executed when the given event occurs

        Args:
            key: Key to refer to field.
            callback: Function to be called when event occurs.
            event: Event to trigger action. Use button release to avoid
              clash with button activate appearance behavior.
        """
        # Tk passes event object to callback which is never used. To avoid all
        # callbacks having to ignore it, the lambda below absorbs & discards it.
        mouse_bind(self.fields[key], event, lambda _event: callback())


class ScrolledReadOnlyText(tk.Text):
    """Implement a read only mode text editor class with scroll bar.

    Done by replacing the bindings for the insert and delete events. From:
    http://stackoverflow.com/questions/3842155/is-there-a-way-to-make-the-tkinter-text-widget-read-only
    """

    # Tag can be used to select a line of text, and to search for the selected line
    # Can't use standard selection since that would interfere with user trying to copy/paste, etc.
    SELECT_TAG_NAME = "chk_select"

    def __init__(self, parent, context_menu=True, **kwargs):  # type: ignore[no-untyped-def]
        """Init the class and set the insert and delete event bindings."""

        self.frame = ttk.Frame(parent)
        self.frame.grid(row=0, column=0, sticky="NSEW")
        self.frame.columnconfigure(0, weight=1)
        self.frame.rowconfigure(0, weight=1)

        super().__init__(self.frame, **kwargs)
        super().grid(column=0, row=0, sticky="NSEW")
        self.redirector = WidgetRedirector(self)
        self.insert = self.redirector.register("insert", lambda *args, **kw: "break")
        self.delete = self.redirector.register("delete", lambda *args, **kw: "break")

        hscroll = ttk.Scrollbar(self.frame, orient=tk.HORIZONTAL, command=self.xview)
        hscroll.grid(column=0, row=1, sticky="NSEW")
        self["xscrollcommand"] = hscroll.set
        vscroll = ttk.Scrollbar(self.frame, orient=tk.VERTICAL, command=self.yview)
        vscroll.grid(column=1, row=0, sticky="NSEW")
        self["yscrollcommand"] = vscroll.set

        self["inactiveselect"] = self["selectbackground"]

        self.tag_configure(
            ScrolledReadOnlyText.SELECT_TAG_NAME,
            background="#dddddd",
            foreground="#000000",
        )

        self["cursor"] = "arrow"

        # Since Text widgets don't normally listen to theme changes,
        # need to do it explicitly here.
        super().bind(
            "<<ThemeChanged>>", lambda _event: theme_set_tk_widget_colors(self)
        )
        # Also on creation, so it's correct for the current theme
        theme_set_tk_widget_colors(self)

        # Redirect attempts to undo & redo to main text window
        # Keystrokes match those in Undo/Redo menu buttons, with case handled manually here
        _, key_event = process_accel("Cmd/Ctrl+Z")
        super().bind(key_event, lambda _event: maintext().event_generate("<<Undo>>"))
        _, key_event = process_accel("Cmd/Ctrl+z")
        super().bind(key_event, lambda _event: maintext().event_generate("<<Undo>>"))
        _, key_event = process_accel("Cmd+Shift+Z" if is_mac() else "Ctrl+Y")
        super().bind(key_event, lambda _event: maintext().event_generate("<<Redo>>"))
        _, key_event = process_accel("Cmd+Shift+z" if is_mac() else "Ctrl+y")
        super().bind(key_event, lambda _event: maintext().event_generate("<<Redo>>"))

        if context_menu:
            add_text_context_menu(self, read_only=True)

    def grid(self, *args: Any, **kwargs: Any) -> None:
        """Override ``grid``, so placing Text actually places surrounding Frame"""
        return self.frame.grid(*args, **kwargs)

    def select_line(self, line_num: int) -> None:
        """Highlight the line_num'th line of text, removing any other highlights.

        Args:
            line_num: Line number to be highlighted - assumed valid.
        """
        self.tag_remove(ScrolledReadOnlyText.SELECT_TAG_NAME, "1.0", tk.END)
        self.tag_add(
            ScrolledReadOnlyText.SELECT_TAG_NAME, f"{line_num}.0", f"{line_num + 1}.0"
        )
        self.see(f"{line_num}.0")

    def get_select_line_num(self) -> Optional[int]:
        """Get the line number of the currently selected line.

        Returns:
            Line number of selected line, or None if no line selected.
        """
        if tag_range := self.tag_nextrange(ScrolledReadOnlyText.SELECT_TAG_NAME, "1.0"):
            return IndexRowCol(tag_range[0]).row
        return None


class MessageLogDialog(ToplevelDialog):
    """A dialog that displays error/info messages."""

    manual_page = "View_Menu#Message_Log"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize messagelog dialog."""
        super().__init__("Message Log", *args, **kwargs)
        self.messagelog = ScrolledReadOnlyText(self.top_frame, wrap=tk.NONE)
        self.messagelog.grid(column=0, row=0, sticky="NSEW")

    def append(self, message: str) -> None:
        """Append a message to the message log dialog."""
        self.messagelog.insert("end", message)


class ErrorHandler(logging.Handler):
    """Handle GUI output of error messages."""

    def __init__(self, *args: Any) -> None:
        """Initialize error logging handler."""
        super().__init__(*args)

    def emit(self, record: logging.LogRecord) -> None:
        """Output error message to message box.

        Args:
            record: Record containing error message.
        """
        messagebox.showerror(title=record.levelname, message=record.getMessage())


class MessageLog(logging.Handler):
    """Handle GUI output of all messages."""

    def __init__(self, *args: Any) -> None:
        """Initialize the message log handler."""
        super().__init__(*args)
        self._messagelog: str = ""
        self.dialog: MessageLogDialog

    def emit(self, record: logging.LogRecord) -> None:
        """Log message in message log.

        Args:
            record: Record containing message.
        """
        message = self.format(record) + "\n"
        self._messagelog += message

        # If dialog is visible, append error
        if hasattr(self, "dialog") and self.dialog.winfo_exists():
            self.dialog.append(message)
            self.dialog.lift()

    def show(self) -> None:
        """Show the message log dialog."""
        already_shown = hasattr(self, "dialog") and self.dialog.winfo_exists()
        self.dialog = MessageLogDialog.show_dialog()
        if not already_shown:
            self.dialog.append(self._messagelog)
        self.dialog.lift()


class MainWindow:
    """Handles the construction of the main window with its basic widgets

    These class variables are set in ``__init__`` to store the single instance
    of these main window items. They are exposed externally via convenience
    functions with the same names, e.g. ``root()`` returns ``MainWindow.root``
    """

    menubar: tk.Menu
    mainimage: MainImage
    statusbar: StatusBar
    messagelog: MessageLog
    busy_widget: ttk.Label

    def __init__(self) -> None:
        Root()
        # Themes
        themed_style(ttk.Style())

        MainWindow.menubar = tk.Menu()
        root()["menu"] = menubar()
        MainWindow.messagelog = MessageLog()

        status_frame = ttk.Frame(root())
        status_frame.grid(
            column=STATUS_COL,
            row=STATUS_ROW,
            sticky="NSEW",
        )
        status_frame.columnconfigure(0, weight=1)
        MainWindow.statusbar = StatusBar(status_frame)
        MainWindow.statusbar.grid(
            column=0,
            row=0,
            sticky="NSW",
        )
        MainWindow.busy_widget = ttk.Label(status_frame, foreground="red")
        MainWindow.busy_widget.grid(
            column=1,
            row=0,
            sticky="NSE",
            padx=10,
        )
        Busy.busy_widget_setup(MainWindow.busy_widget)

        ttk.Separator(root()).grid(
            column=SEPARATOR_COL,
            row=SEPARATOR_ROW,
            sticky="NSEW",
        )

        self.paned_window = tk.PanedWindow(
            root(),
            orient=tk.HORIZONTAL,
            sashwidth=4,
            sashrelief=tk.RIDGE,
            showhandle=True,
            handlesize=10,
        )
        self.paned_window.grid(
            column=TEXTIMAGE_WINDOW_COL, row=TEXTIMAGE_WINDOW_ROW, sticky="NSEW"
        )

        self.paned_text_window = tk.PanedWindow(
            self.paned_window,
            orient=tk.VERTICAL,
            sashwidth=4,
            sashrelief=tk.RIDGE,
            showhandle=True,
            handlesize=10,
        )
        self.paned_text_window.grid(
            column=TEXTIMAGE_WINDOW_COL, row=TEXTIMAGE_WINDOW_ROW, sticky="NSEW"
        )

        MainText(
            self.paned_text_window,
            root(),
            undo=True,
            wrap="none",
            autoseparators=True,
            maxundo=-1,
            highlightthickness=2,
        )

        self.paned_window.add(self.paned_text_window, minsize=MIN_PANE_WIDTH)
        add_text_context_menu(maintext())
        add_text_context_menu(maintext().peer)

        MainWindow.mainimage = MainImage(
            self.paned_window,
            hide_func=self.hide_image,
            float_func=self.float_image,
            dock_func=self.dock_image,
        )

    def hide_image(self) -> None:
        """Stop showing the current image."""
        preferences.set(PrefKey.AUTO_IMAGE, False)
        self.clear_image()
        root().wm_forget(mainimage())  # type: ignore[arg-type]
        self.paned_window.forget(mainimage())

    def float_image(self, _event: Optional[tk.Event] = None) -> None:
        """Float the image into a separate window"""
        mainimage().grid_remove()
        if mainimage().is_image_loaded():
            root().wm_manage(mainimage())
            mainimage().lift()
            # Obscure tk.Wm calls needed because although mainimage has been converted
            # to a toplevel by Tk, it doesn't appear as though tkinter knows about it,
            # so can't call mainimage().wm_geometry() or set the size via normal config
            # methods.
            tk.Wm.geometry(mainimage(), preferences.get(PrefKey.IMAGE_FLOAT_GEOMETRY))  # type: ignore[call-overload]
            tk.Wm.protocol(mainimage(), "WM_DELETE_WINDOW", self.hide_image)  # type: ignore[call-overload]
        else:
            root().wm_forget(mainimage())  # type: ignore[arg-type]
        preferences.set(PrefKey.IMAGE_WINDOW_DOCKED, False)

        # It is OK to save image viewer geometry from now on
        mainimage().enable_geometry_storage()

    def dock_image(self, _event: Optional[tk.Event] = None) -> None:
        """Dock the image back into the main window"""
        root().wm_forget(mainimage())  # type: ignore[arg-type]
        if mainimage().is_image_loaded():
            self.paned_window.add(mainimage(), minsize=MIN_PANE_WIDTH)
            self.paned_window.sash_place(
                0, preferences.get(PrefKey.IMAGE_DOCK_SASH_COORD), 0
            )
        else:
            try:
                self.paned_window.forget(mainimage())
            except tk.TclError:
                pass  # OK - image wasn't being managed by paned_window
        preferences.set(PrefKey.IMAGE_WINDOW_DOCKED, True)

    def load_image(self, filename: str) -> None:
        """Load the image for the given page.

        Args:
            filename: Path to image file.
            force_show: True to force dock or float image
        """
        image_already_loaded = mainimage().is_image_loaded()
        if mainimage().load_image(filename) and not image_already_loaded:
            if preferences.get(PrefKey.IMAGE_WINDOW_DOCKED):
                self.dock_image()
            else:
                self.float_image()

    def clear_image(self) -> None:
        """Clear the image currently being shown."""
        mainimage().clear_image()


def do_sound_bell() -> None:
    """Sound warning bell audibly and/or visually.

    Audible uses the default system bell sound.
    Visible flashes the first statusbar button (must be ttk.Button)
    """
    if preferences.get(PrefKey.BELL_AUDIBLE):
        root().bell()
    if preferences.get(PrefKey.BELL_VISUAL):
        bell_button = statusbar().fields["rowcol"]
        # Belt & suspenders: uses the "disabled" state of button in temporary style,
        # but also restores setting in temporary style, and restores default style.
        style = ttk.Style()
        # Set temporary style's disabled bg to red
        style.map("W.TButton", foreground=[("disabled", "red")])
        # Save current disabled bg default for buttons
        save_bg = style.lookup("TButton", "background", state=[("disabled")])
        # Save style currently used by button
        cur_style = bell_button["style"]
        # Set button to use temporary style
        bell_button.configure(style="W.TButton")
        # Flash 3 times
        for state in ("disabled", "normal", "disabled", "normal", "disabled", "normal"):
            bell_button["state"] = state
            bell_button.update()
            time.sleep(0.08)
        # Set button to use its previous style again
        bell_button.configure(style=cur_style)
        # Just in case, set the temporary style back to the default
        style.map("W.TButton", background=[("disabled", save_bg)])


bell_set_callback(do_sound_bell)


def add_text_context_menu(text_widget: tk.Text, read_only: bool = False) -> None:
    """Add a context menu to a Text widget.

    Puts Cut, Copy, Paste, Select All menu buttons in a context menu.

    Args:
        read_only: True if text is read-only, so does not require Cut & Paste options.
    """
    menu_context = Menu(text_widget, "")
    menu_context.add_cut_copy_paste(read_only=read_only)

    def post_context_menu(event: tk.Event) -> None:
        event.widget.focus_set()
        menu_context.post(event.x_root, event.y_root)

    mouse_bind(text_widget, "3", post_context_menu)


def mainimage() -> MainImage:
    """Return the single MainImage widget"""
    assert MainWindow.mainimage is not None
    return MainWindow.mainimage


def menubar() -> tk.Menu:
    """Return the single Menu widget used as the menubar"""
    assert MainWindow.menubar is not None
    return MainWindow.menubar


def statusbar() -> StatusBar:
    """Return the single StatusBar widget"""
    assert MainWindow.statusbar is not None
    return MainWindow.statusbar



"""Tools relating to HTML."""

import gzip
from html.parser import HTMLParser
import logging
import os.path
import tkinter as tk
from tkinter import ttk, filedialog
from typing import Optional, Any

from PIL import Image, ImageTk, UnidentifiedImageError
import regex as re
import requests

from guiguts.checkers import CheckerDialog
from guiguts.file import the_file
from guiguts.maintext import maintext
from guiguts.preferences import (
    PersistentString,
    PersistentBoolean,
    PrefKey,
    preferences,
)
from guiguts.utilities import IndexRange, sound_bell, DiacriticRemover, IndexRowCol
from guiguts.widgets import ToplevelDialog

logger = logging.getLogger(__package__)

IMAGE_THUMB_SIZE = 200
RETURN_ARROW = "⏎"
EM_PX = 16.0  # 16 pixels per em
LAND_X = 4  # Common landscape screen ratio
LAND_Y = 3


class HTMLImageDialog(ToplevelDialog):
    """Dialog for inserting image markup into HTML."""

    manual_page = "HTML_Menu#Add_Illustrations"

    def __init__(self) -> None:
        """Initialize HTML Image dialog."""
        super().__init__(
            "HTML Images",
            resize_x=False,
            resize_y=False,
        )

        self.image: Optional[Image.Image] = None
        self.imagetk: Optional[ImageTk.PhotoImage] = None
        self.width = 0
        self.height = 0
        self.illo_range: Optional[IndexRange] = None

        # File
        file_frame = ttk.LabelFrame(self.top_frame, text="File", padding=2)
        file_frame.grid(row=0, column=0, sticky="NSEW")
        file_frame.columnconfigure(0, weight=1)
        file_name_frame = ttk.Frame(file_frame)
        file_name_frame.grid(row=0, column=0)
        file_name_frame.columnconfigure(0, weight=1)
        self.filename_textvariable = tk.StringVar(self, "")
        self.fn_entry = ttk.Entry(
            file_name_frame,
            textvariable=self.filename_textvariable,
            width=30,
        )
        self.fn_entry.grid(row=0, column=0, sticky="EW", padx=(0, 2))
        ttk.Button(
            file_name_frame,
            text="Browse...",
            command=self.choose_file,
            takefocus=False,
        ).grid(row=0, column=1, sticky="NSEW")

        # Buttons to see prev/next file
        file_btn_frame = ttk.Frame(file_frame)
        file_btn_frame.grid(row=1, column=0)
        ttk.Button(
            file_btn_frame,
            text="Prev File",
            command=lambda: self.next_file(reverse=True),
            takefocus=False,
        ).grid(row=0, column=0, padx=2)
        ttk.Button(
            file_btn_frame,
            text="Next File",
            command=lambda: self.next_file(reverse=False),
            takefocus=False,
        ).grid(row=0, column=1, padx=2)

        # Label to display thumbnail of image - allocate a
        # square space the same width as the filename frame
        file_name_frame.update_idletasks()
        frame_width = file_name_frame.winfo_width()
        thumbnail_frame = ttk.LabelFrame(self.top_frame, text="Thumbnail")
        thumbnail_frame.grid(row=1, column=0, sticky="NSEW")
        thumbnail_frame.columnconfigure(0, minsize=frame_width, weight=1)
        thumbnail_frame.rowconfigure(0, minsize=frame_width, weight=1)
        self.thumbnail = ttk.Label(thumbnail_frame, justify=tk.CENTER)
        self.thumbnail.grid(row=0, column=0)
        self.thumbsize = frame_width - 10

        # Caption text
        caption_frame = ttk.LabelFrame(self.top_frame, text="Caption text", padding=2)
        caption_frame.grid(row=2, column=0, sticky="NSEW")
        caption_frame.columnconfigure(0, weight=1)
        self.caption_textvariable = tk.StringVar(self, "")
        ttk.Entry(
            caption_frame,
            textvariable=self.caption_textvariable,
        ).grid(row=0, column=0, sticky="NSEW")

        # Alt text
        alt_frame = ttk.LabelFrame(self.top_frame, text="Alt text", padding=2)
        alt_frame.grid(row=3, column=0, sticky="NSEW")
        alt_frame.columnconfigure(0, weight=1)
        self.alt_textvariable = tk.StringVar(self, "")
        ttk.Entry(
            alt_frame,
            textvariable=self.alt_textvariable,
        ).grid(row=0, column=0, sticky="NSEW")

        # Geometry
        geom_frame = ttk.LabelFrame(self.top_frame, padding=2, text="Geometry")
        geom_frame.grid(row=4, column=0, pady=(5, 0), sticky="NSEW")
        geom_frame.columnconfigure(0, weight=1)
        width_height_frame = ttk.Frame(geom_frame)
        width_height_frame.grid(row=0, column=0)
        width_height_frame.columnconfigure(0, weight=1)
        self.image_width = 0
        self.image_height = 0

        def width_updated(new_value: str) -> bool:
            """Use validation routine for width box to update height box."""
            self.height_textvariable.set(self.get_height_from_width(new_value))
            return True

        ttk.Label(width_height_frame, text="Width").grid(row=0, column=0, padx=4)
        self.width_textvariable = tk.StringVar(self, "")
        ttk.Entry(
            width_height_frame,
            textvariable=self.width_textvariable,
            width=8,
            validate="all",
            validatecommand=(self.register(width_updated), "%P"),
        ).grid(row=0, column=1, sticky="NSEW", padx=(4, 10))
        ttk.Label(width_height_frame, text="Height").grid(row=0, column=2, padx=(10, 4))
        self.height_textvariable = tk.StringVar(self, "")
        ttk.Entry(
            width_height_frame,
            textvariable=self.height_textvariable,
            width=8,
            state=tk.DISABLED,
        ).grid(row=0, column=3, sticky="NSEW", padx=4)

        unit_frame = ttk.Frame(geom_frame)
        unit_frame.grid(row=1, column=0, pady=5)
        unit_textvariable = PersistentString(PrefKey.HTML_IMAGE_UNIT)
        ttk.Radiobutton(
            unit_frame,
            text="%",
            variable=unit_textvariable,
            value="%",
            takefocus=False,
            command=self.update_geometry_fields,
        ).grid(row=0, column=4, sticky="NSEW", padx=10)
        ttk.Radiobutton(
            unit_frame,
            text="em",
            variable=unit_textvariable,
            value="em",
            takefocus=False,
            command=self.update_geometry_fields,
        ).grid(row=0, column=5, sticky="NSEW", padx=10)
        ttk.Radiobutton(
            unit_frame,
            text="px",
            variable=unit_textvariable,
            value="px",
            takefocus=False,
            command=self.update_geometry_fields,
        ).grid(row=0, column=6, sticky="NSEW", padx=10)

        self.file_info_textvariable = tk.StringVar(self, "")
        ttk.Label(geom_frame, text="", textvariable=self.file_info_textvariable).grid(
            row=2, column=0
        )
        self.max_width_textvariable = tk.StringVar(self, "")
        ttk.Label(geom_frame, text="", textvariable=self.max_width_textvariable).grid(
            row=3, column=0
        )
        self.override_checkbutton = ttk.Checkbutton(
            geom_frame,
            text="Override % with 100% in epub",
            variable=PersistentBoolean(PrefKey.HTML_IMAGE_OVERRIDE_EPUB),
            takefocus=False,
        )
        self.override_checkbutton.grid(row=4, column=0, sticky="NS")
        # Dummy placeholder label for when above checkbutton is ungridded
        ttk.Label(geom_frame, width=0, text="").grid(
            row=4, column=1, sticky="NS", pady=3
        )

        # Alignment
        align_frame = ttk.LabelFrame(self.top_frame, padding=2, text="Alignment")
        align_frame.grid(row=5, column=0, pady=(5, 0), sticky="NSEW")
        align_frame.columnconfigure(0, weight=1)
        align_frame.columnconfigure(1, weight=1)
        align_frame.columnconfigure(2, weight=1)
        align_textvariable = PersistentString(PrefKey.HTML_IMAGE_ALIGNMENT)
        ttk.Radiobutton(
            align_frame,
            text="Left",
            variable=align_textvariable,
            value="left",
            takefocus=False,
        ).grid(row=0, column=0, sticky="NSW", padx=10)
        ttk.Radiobutton(
            align_frame,
            text="Center",
            variable=align_textvariable,
            value="center",
            takefocus=False,
        ).grid(row=0, column=1, sticky="NSW", padx=10)
        ttk.Radiobutton(
            align_frame,
            text="Right",
            variable=align_textvariable,
            value="right",
            takefocus=False,
        ).grid(row=0, column=2, sticky="NSW", padx=10)

        # Buttons to Find illos and Convert to HTML
        btn_frame = ttk.Frame(self.top_frame, padding=2)
        btn_frame.grid(row=6, column=0, pady=(5, 0))
        ttk.Button(
            btn_frame,
            text="Convert to HTML",
            command=self.convert_to_html,
            takefocus=False,
            width=18,
        ).grid(row=0, column=0, sticky="NSEW", padx=2)
        ttk.Button(
            btn_frame,
            text="Find [Illustration]",
            command=self.find_illo_markup,
            takefocus=False,
            width=18,
        ).grid(row=0, column=1, sticky="NSEW", padx=2)

        self.find_illo_markup()

    def load_file(self, file_name: str) -> None:
        """Load given image file."""
        assert file_name
        file_name = os.path.normpath(file_name)
        if not os.path.isfile(file_name):
            logger.error(f"Unsuitable image file: {file_name}")
            self.clear_image()
            return

        # Display filename and load image file
        rel_file_name = os.path.relpath(
            file_name, start=os.path.dirname(the_file().filename)
        )
        # We want forward slashes in HTML file, even if Windows uses backslashes
        self.filename_textvariable.set(rel_file_name.replace("\\", "/"))
        self.fn_entry.xview_moveto(1.0)
        if self.image is not None:
            del self.image
        try:
            self.image = Image.open(file_name).convert("RGB")
        except UnidentifiedImageError:
            self.image = None
            logger.error(f"Unable to identify image file: {file_name}")
            return
        self.image_width, self.image_height = self.image.size
        self.update_geometry_fields()

        # Resize image to fit thumbnail label
        scale = min(
            self.thumbsize / self.image_width, self.thumbsize / self.image_height, 1.0
        )
        width = int(self.image_width * scale)
        height = int(self.image_height * scale)
        image = self.image.resize(
            size=(width, height), resample=Image.Resampling.LANCZOS
        )
        if self.imagetk:
            del self.imagetk
        self.imagetk = ImageTk.PhotoImage(image)
        del image
        self.thumbnail.config(image=self.imagetk)
        em_info = f"{self.image_width / EM_PX:.4f} x {self.image_height / EM_PX:.4f} em"
        px_info = f"({self.image_width} x {self.image_height} px)"
        self.file_info_textvariable.set(f"File size: {em_info} {px_info}")
        self.lift()

    def reset(self) -> None:
        """Reset dialog, removing spotlights."""
        maintext().remove_spotlights()

    def choose_file(self) -> None:
        """Allow user to choose image file."""
        if file_name := filedialog.askopenfilename(
            filetypes=(
                ("Image files", "*.jpg *.png *.gif"),
                ("All files", "*.*"),
            ),
            title="Select Image File",
            parent=self,
        ):
            self.load_file(file_name)

    def next_file(self, reverse: bool = False) -> None:
        """Load the next file alphabetically.

        Args:
            reverse: True to load previous file instead.
        """
        # If no current file, can't get "next", so make user choose one
        current_fn = self.filename_textvariable.get()
        if not current_fn:
            self.choose_file()
            return
        current_fn = os.path.join(os.path.dirname(the_file().filename), current_fn)
        # Check current directory is valid
        current_dir = os.path.dirname(current_fn)
        if not os.path.isdir(current_dir):
            logger.error(f"Image directory invalid: {current_dir}")
            return
        current_basename = os.path.basename(current_fn)
        found = False
        for fn in sorted(os.listdir(current_dir), reverse=reverse):
            # Skip non-image files by checking extension
            if os.path.splitext(fn)[1] not in (".jpg", ".gif", ".png"):
                continue
            # If found on previous time through loop, this is the file we want
            if found:
                self.load_file(os.path.join(current_dir, fn))
                return
            if fn == current_basename:
                found = True
        # Reached end of dir listing without finding next file
        sound_bell()

    def find_illo_markup(self) -> None:
        """Find first unconverted illo markup in file and
        advance to the next file."""
        self.illo_range = None
        # Find and go to start of first unconverted illo markup
        illo_match_start = maintext().find_match(
            r"(<p>)?\[Illustration",
            IndexRange(maintext().start(), maintext().end()),
            regexp=True,
        )
        if illo_match_start is None:
            sound_bell()
            return
        maintext().set_insert_index(illo_match_start.rowcol, focus=False)
        # Find end of markup and spotlight it
        search_start = maintext().rowcol(f"{illo_match_start.rowcol.index()}+12c")
        nested = False
        while True:
            illo_match_end = maintext().find_match(
                r"[][]",
                IndexRange(search_start, maintext().end()),
                regexp=True,
            )
            if illo_match_end is None:
                break
            start_index = illo_match_end.rowcol.index()
            match_text = maintext().get(start_index, f"{start_index} lineend")
            # Have we found the end of the illo markup (i.e. end of line bracket)
            if not nested and re.fullmatch("] *$", match_text):
                break
            # Keep track of whether there are nested brackets, e.g. [12] inside illo markup
            nested = match_text[0] == "["
            search_start = maintext().rowcol(f"{illo_match_end.rowcol.index()}+1c")
        if illo_match_end is None:
            logger.error("Unclosed [Illustration markup")
            return
        self.illo_range = IndexRange(
            illo_match_start.rowcol,
            maintext().rowcol(
                f"{illo_match_end.rowcol.index()}+{illo_match_end.count}c"
            ),
        )
        maintext().spotlight_range(self.illo_range)
        # Display caption in dialog and add <p> markup if none
        caption = maintext().get(
            self.illo_range.start.index(), self.illo_range.end.index()
        )
        caption = re.sub(r"(?<=(^<p.?>|^))\[Illustration:? ?", "", caption)
        caption = re.sub(r"\](?=(</p> *$| *$))", "", caption)
        # Remove simple <p> markup and replace newlines with return_arrow character
        if caption:
            caption = re.sub("</?p>", "", caption)
            caption = re.sub("^\n+", "", caption)
            caption = re.sub("\n$", "", caption)
            caption = re.sub("\n", RETURN_ARROW, caption)
        self.caption_textvariable.set(caption)
        # Clear alt text, ready for user to type in required string
        self.alt_textvariable.set("")
        self.next_file()
        self.lift()

    def convert_to_html(self) -> None:
        """Convert selected [Illustration...] markup to HTML."""
        filename = self.filename_textvariable.get()
        if self.illo_range is None or not filename:
            sound_bell()
            return
        # Get caption & add some space to prettify HTML
        caption = self.caption_textvariable.get()
        if caption:
            caption = re.sub(RETURN_ARROW, "\n    ", f"      {caption}")
            caption = f"  <figcaption>\n{caption}\n  </figcaption>\n"
        # Now alt text - escape any double quotes
        alt = self.alt_textvariable.get().replace('"', "&quot;")
        alt = f' alt="{alt}"'
        # Create a unique ID from the filename
        image_id = os.path.splitext(os.path.basename(filename))[0]
        image_id = DiacriticRemover.remove_diacritics(image_id)
        # If ID already exists in file, try suffixes "_2", "_3", etc.
        id_base = image_id
        id_suffix = 1
        whole_file = IndexRange(maintext().start(), maintext().end())
        # Loop until we find an id that is not found in the file
        while maintext().find_match(f'id="{image_id}"', whole_file):
            id_suffix += 1
            image_id = f"{id_base}_{id_suffix}"
        # Alignment
        alignment = f"fig{preferences.get(PrefKey.HTML_IMAGE_ALIGNMENT)}"
        # Set up classes & styles
        unit_type = preferences.get(PrefKey.HTML_IMAGE_UNIT)
        width = self.width_textvariable.get()
        img_size = img_class = fig_class = style = ""
        if unit_type == "px":
            img_size = f' width="{self.image_width}" height="{self.image_height}"'
            style = f' style="width: {self.image_width}px;"'
        else:
            img_class = ' class="w100"'
            fig_class = " illow"
            fig_class += "p" if unit_type == "%" else "e"
            fig_class += width.replace(".", "_")
            if unit_type == "%":  # Never want %-width image to exceed natural size
                style = f' style="max-width: {self.image_width / EM_PX}em;"'

        # Construct HTML
        html = f'<figure class="{alignment}{fig_class}" id="{image_id}"{style}>\n'
        html += f'  <img{img_class} src="{filename}"{img_size}{alt}>\n'
        html += f"{caption}</figure>"

        # Replace [Illustration...] with HTML
        maintext().undo_block_begin()
        maintext().replace(
            self.illo_range.start.index(), self.illo_range.end.index(), html
        )
        self.illo_range = None

        # Now to insert CSS at end of style block, except for px sizes
        insert_point = maintext().search("</style", "1.0", tk.END)
        if unit_type == "px" or not insert_point:
            return
        fig_class = fig_class[1:]  # Remove leading space to get classname
        class_def = f".{fig_class} {{width: {width}{unit_type};}}"
        cssdef = class_def
        # If % width and override flag set  then also add CSS to override width to 100% for epub
        if (
            unit_type == "%"
            and preferences.get(PrefKey.HTML_IMAGE_OVERRIDE_EPUB)
            and width != "100"
        ):
            cssdef += f"\n.x-ebookmaker .{fig_class} {{width: 100%;}}"
        # Add heading if there's not one already
        heading = "/* Illustration classes */"
        if not maintext().search(heading, "1.0", insert_point):
            cssdef = f"\n{heading}\n{cssdef}"
        # Only insert if definition not already in file
        if not maintext().search(class_def, "1.0", insert_point):
            maintext().insert(f"{insert_point} linestart", f"{cssdef}\n")
        # Remove spotlight
        maintext().remove_spotlights()

    def clear_image(self) -> None:
        """Clear the image and reset variables accordingly."""
        if self.image:
            del self.image
        self.image = None
        if self.imagetk:
            del self.imagetk
        self.imagetk = None
        self.thumbnail.config(image="")

    def update_geometry_fields(self) -> None:
        """Update the width and height fields with data from the image."""
        if self.image_width <= 0 or self.image_height <= 0:
            logger.error("Image file has illegal width/height")
            return
        unit_type = preferences.get(PrefKey.HTML_IMAGE_UNIT)
        match unit_type:
            case "%":
                # Percentage width for the current image such that
                # both its width and height will fit a landscape screen
                size_x = f"{min(100, int(100.0 * LAND_Y / LAND_X * self.image_width / self.image_height))}"
            case "em":
                size_x = f"{self.image_width / EM_PX:.4f}"
            case "px":
                size_x = f"{self.image_width}"
            case _:
                size_x = ""
        self.width_textvariable.set(size_x)
        self.height_textvariable.set(self.get_height_from_width(size_x))
        if unit_type == "%":
            # Tell user maximum % width such that both dimensions will fit a 4:3 screen
            self.max_width_textvariable.set(
                f"Max width to fit {LAND_X}:{LAND_Y} screen is {size_x}%"
            )
            self.override_checkbutton.grid()
        else:
            self.max_width_textvariable.set("")
            self.override_checkbutton.grid_remove()

    def get_height_from_width(self, width: str) -> str:
        """Return HTML image height as string, based on width,
        type of units, and aspect ratio of image.

        Args:
            width: HTML width of image as string.

        Return:
            String containing expected height of image (or "--")
        """
        try:
            width_fl = float(width)
        except ValueError:
            return "--"
        if self.image_width <= 0 or self.image_height <= 0:
            return "--"
        match preferences.get(PrefKey.HTML_IMAGE_UNIT):
            case "%":
                return "--"
            case "em":
                return f"{width_fl * self.image_height / self.image_width:.4f}"
            case "px":
                return f"{width_fl * self.image_height / self.image_width:.0f}"
        return "--"


def html_validator_check() -> None:
    """Validate the current HTML file."""

    class HTMLValidatorDialog(CheckerDialog):
        """Minimal class to identify dialog type so that it can exist
        simultaneously with other checker dialogs."""

        manual_page = "HTML_Menu#HTML_Link_Checker"

        def __init__(self, **kwargs: Any) -> None:
            """Initialize HTML Validator dialog."""

            super().__init__(
                "HTML Validator Results",
                tooltip="\n".join(
                    [
                        "Left click: Select & find validation error",
                        "Right click: Remove validation error from this list",
                    ]
                ),
                **kwargs,
            )

    checker_dialog = HTMLValidatorDialog.show_dialog(rerun_command=html_validator_check)

    do_validator_check(checker_dialog)


def do_validator_check(checker_dialog: CheckerDialog) -> None:
    """Do the actual check and add messages to the dialog."""

    validator_url = "https://validator.w3.org/nu/"

    def report_exception(message: str, exc: Exception | str) -> None:
        """Report exception to user and suggest manual validation.
        Also call `display_entries to clear "busy" message.

        Args:
            message: Initial part of message to user.
            exc: Exception that was thrown, or a string describing the problem
        """
        logger.error(f"{message}\nValidate manually online.\nError details:\n{exc}")
        checker_dialog.display_entries()

    try:
        req = requests.post(
            validator_url,
            data=gzip.compress(bytes(maintext().get("1.0", tk.END), "UTF-8")),
            params={"out": "json"},
            headers={
                "Content-Type": "text/html; charset=UTF-8",
                "Content-Encoding": "gzip",
                "Accept-Encoding": "gzip",
            },
            timeout=15,
        )
    except requests.exceptions.Timeout as exc:
        report_exception(f"Request to {validator_url} timed out.", exc)
        return
    except ConnectionError as exc:
        report_exception(f"Connection error to {validator_url}.", exc)
        return
    # Check if HTTP request was unsuccessful
    try:
        req.raise_for_status()
    except (requests.exceptions.HTTPError, requests.exceptions.TooManyRedirects) as exc:
        report_exception(f"Request to {validator_url} was unsuccessful.", exc)
        return

    # Even if there are no errors, there should still be a messages list
    try:
        messages = req.json()["messages"]
    except KeyError:
        report_exception(
            f"Invalid data returned from {validator_url}.", 'No "messages" key'
        )
        return

    # Add a line to the checker dialog for each message
    for message in messages:
        try:
            end = IndexRowCol(int(message["lastLine"]), int(message["lastColumn"]))
        except KeyError:
            end = None
        # Missing start line means it's all on one line, so same as end line
        try:
            start_row = int(message["firstLine"])
        except KeyError:
            if end is None:
                start_row = None
            else:
                start_row = end.row
        # Missing start column means it's just one character
        try:
            start_col = int(message["firstColumn"]) - 1
        except KeyError:
            if end is None:
                start_col = None
            else:
                start_col = end.col - 1
        if start_row is None or start_col is None:
            start = None
        else:
            # Messages sometimes range from end of previous line,
            # in which case switch to start of next line
            if start_col >= maintext().rowcol(f"{start_row}.end").col:
                start_col = 0
                start_row += 1
            start = IndexRowCol(start_row, start_col)
        if start is None or end is None:
            error_range = None
        else:
            error_range = IndexRange(start, end)
        try:
            line = message["message"]
        except KeyError:
            line = "Data error - no message found"
        try:
            error_type = f'{message["type"].upper()}: '
        except KeyError:
            error_type = ""

        checker_dialog.add_entry(line, error_range, error_prefix=error_type)

    if not messages:
        checker_dialog.add_entry("No errors reported by validator")
    checker_dialog.display_entries()


def html_link_check() -> None:
    """Validate the current HTML file."""

    class HTMLLinkCheckerDialog(CheckerDialog):
        """HTML Link Checker dialog."""

        manual_page = "HTML_Menu#HTML_Link_Checker"

        def __init__(self, **kwargs: Any) -> None:
            """Initialize HTML Link Checker dialog."""

            super().__init__(
                "HTML Link Checker Results",
                tooltip="\n".join(
                    [
                        "Left click: Select & find issue",
                        "Right click: Remove issue from this list",
                    ]
                ),
                **kwargs,
            )

    checker_dialog = HTMLLinkCheckerDialog.show_dialog(rerun_command=html_link_check)

    do_link_check(checker_dialog)


def do_link_check(checker_dialog: CheckerDialog) -> None:
    """Do the actual check and add messages to the dialog."""

    class AttrPos:
        """Class to store attribute & position in file."""

        def __init__(self, attr: str, value: str, position: tuple[int, int]) -> None:
            """Initialize attribute pos class

            Args:
                attr: Attribute name, e.g. "href".
                value: Reference to file or location, e.g. "images/i1.jpg" or "#page1".
                position: Line & column number of start of attribute in file.
            """
            self.attr = attr
            self.value = value
            self.rowcol = IndexRowCol(position[0], position[1])

    class UrlPos:
        """Class to store url & position in file."""

        def __init__(self, value: str, position: tuple[int, int], count: int) -> None:
            """Initialize url pos class

            Args:
                value: File name, e.g. "images/i1.jpg".
                position: Line & column number of start of `style` attribute in file.
                count: Which url("...") within `style` attribute
            """
            self.value = value
            self.rowcol = IndexRowCol(position[0], position[1])
            self.count = count

    class IdPos:
        """Class to store position of id in file and whether used.
        Name of id is used as key in dict.
        """

        def __init__(self, position: tuple[int, int]) -> None:
            """Initialize id pos class.

            Args:
                position: Line & column number of start of `id` in file.
            """
            self.rowcol = IndexRowCol(position[0], position[1])
            self.used = False

    links: list[AttrPos] = []
    ids: dict[str, IdPos] = {}  # Dict for speed of lookup
    urls: list[UrlPos] = []

    class HTMLParserLink(HTMLParser):
        """Class to parse HTML."""

        def handle_starttag(
            self, tag: str, attrs: list[tuple[str, str | None]]
        ) -> None:
            """Handle an HTML start tag"""
            for attr in attrs:
                match attr:
                    case ("href" | "src", value) if value is not None:
                        links.append(AttrPos(attr[0], value, self.getpos()))
                    case ("id", value) if value is not None:
                        ids[value] = IdPos(self.getpos())
                    case ("style", value) if value is not None:
                        for num, match in enumerate(
                            re.finditer(r"""\burl\(['"](.*?)['"]\)""", value)
                        ):
                            urls.append(UrlPos(match[1], self.getpos(), num))

    def get_index_range(link: AttrPos) -> IndexRange:
        """Get relevant index range, given info about tag & attribute.

        Args:
            link: Info about tag and attribute.
        """
        rgx = rf"""{link.attr} *= *["'][^'"]*["']"""
        length = tk.IntVar()
        if attr_start := maintext().search(
            rgx, link.rowcol.index(), tk.END, regexp=True, count=length
        ):
            return IndexRange(
                attr_start, maintext().index(f"{attr_start}+{length.get()}c")
            )
        return IndexRange(link.rowcol, link.rowcol)

    def get_url_range(url: UrlPos) -> IndexRange:
        """Get relevant index range, given info about tag & attribute.

        Args:
            url: Info about tag and attribute.
        """
        rgx = r"""url\(['"][^'"]*['"]\)"""
        length = tk.IntVar()
        count = 0
        url_start = url.rowcol.index()
        while url_start := maintext().search(
            rgx, f"{url_start}+5c", tk.END, regexp=True, count=length
        ):
            if count == url.count:
                return IndexRange(
                    url_start, maintext().index(f"{url_start}+{length.get()}c")
                )
            count += 1
        return IndexRange(url.rowcol, url.rowcol)

    # Get list of image files to check if they are referenced
    cur_dir = os.path.dirname(the_file().filename)
    images_used: dict[str, bool] = {}
    images_dir = os.path.join(cur_dir, "images")
    if not os.path.isdir(images_dir):
        logger.error(f"Directory {images_dir} does not exist")
        checker_dialog.display_entries()
        return
    for fn in os.listdir(images_dir):
        # Force forward slash (unlike os.join) so it matches attribute value
        images_used[f"images/{fn}"] = False

    # Parse HTML - tags trigger calls to handle_starttag above
    parser = HTMLParserLink()
    parser.feed(maintext().get_text())

    n_externals = 0
    # Report on any broken links
    for link in links:
        if not link.value.strip():
            checker_dialog.add_entry(f"Empty {link.attr} string", get_index_range(link))
        elif re.match("https?:", link.value):
            checker_dialog.add_entry(
                f"External link: {link.value}", get_index_range(link)
            )
            n_externals += 1
        elif link.value.startswith("#"):
            if link.value[1:] in ids:
                ids[link.value[1:]].used = True
            else:
                checker_dialog.add_entry(
                    f"Internal link without anchor: {link.value}",
                    get_index_range(link),
                )
        else:
            if any(char.isupper() for char in link.value):
                checker_dialog.add_entry(
                    f"Filename contains uppercase: {link.value}", get_index_range(link)
                )
            if os.path.isfile(os.path.join(cur_dir, link.value)):
                images_used[link.value] = True
            else:
                checker_dialog.add_entry(
                    f"File not found: {link.value}", get_index_range(link)
                )

    # Report on any broken urls
    for url in urls:
        if not url.value.strip():
            checker_dialog.add_entry("Empty url string", get_url_range(url))
        elif re.match("https?:", url.value):
            checker_dialog.add_entry(
                f"External url link: {url.value}", get_url_range(url)
            )
            n_externals += 1
        else:
            if any(char.isupper() for char in url.value):
                checker_dialog.add_entry(
                    f"Filename contains uppercase: {url.value}", get_url_range(url)
                )
            if os.path.isfile(os.path.join(cur_dir, url.value)):
                images_used[url.value] = True
            else:
                checker_dialog.add_entry(
                    f"Url not found: {url.value}", get_url_range(url)
                )

    # Report any unused image files
    unused_header = False
    for image, used in images_used.items():
        if not used:
            if not unused_header:
                checker_dialog.add_header("")
                checker_dialog.add_header("UNUSED IMAGE FILES")
                unused_header = True
            checker_dialog.add_entry(f"File not used: {image}")

    # Statistics summary
    checker_dialog.add_header("")
    checker_dialog.add_header("LINK STATISTICS")
    checker_dialog.add_entry(f"{len(ids)} anchors (tags with id attribute)")
    n_refs = sum(1 for link in links if link.value.startswith("#"))
    checker_dialog.add_entry(f"{n_refs} internal links (using href)")
    checker_dialog.add_entry(f"{len(links) - n_refs} file links (using href or src)")
    checker_dialog.add_entry(f"{len(urls)} url links (using CSS style url)")

    # Report any unused anchors - last because only informational and may be long
    n_unused = sum(1 for id_pos in ids.values() if not id_pos.used)
    unused_header = False
    rgx = r"""id *= *["'][^'"]*["']"""
    for id_name, id_pos in ids.items():
        if id_pos.used:
            continue
        if not unused_header:
            checker_dialog.add_header("")
            checker_dialog.add_header(
                f"{n_unused} ANCHORS WITHOUT LINKS (INFORMATIONAL)"
            )
            unused_header = True
        length = tk.IntVar()
        if attr_start := maintext().search(
            rgx, id_pos.rowcol.index(), tk.END, regexp=True, count=length
        ):
            checker_dialog.add_entry(
                f"Anchor not used: {id_name}",
                IndexRange(
                    attr_start, maintext().index(f"{attr_start}+{length.get()}c")
                ),
            )
        else:
            checker_dialog.add_entry(f"Anchor not used: {id_name}")

    checker_dialog.display_entries()



"""Handle preferences"""

import copy
from enum import StrEnum, auto
import json
import logging
import os
import tkinter as tk
from typing import Any, Callable

from guiguts.utilities import is_x11, is_windows, called_from_test, load_dict_from_json

if is_windows():
    from win32com.shell import shell, shellcon  # type: ignore # pylint: disable=import-error, no-name-in-module


logger = logging.getLogger(__package__)


class PrefKey(StrEnum):
    """Enum class to store preferences keys."""

    AUTO_IMAGE = auto()
    BELL_AUDIBLE = auto()
    BELL_VISUAL = auto()
    IMAGE_WINDOW_DOCKED = auto()
    RECENT_FILES = auto()
    LINE_NUMBERS = auto()
    ORDINAL_NAMES = auto()
    SEARCH_HISTORY = auto()
    REPLACE_HISTORY = auto()
    SEARCHDIALOG_REVERSE = auto()
    SEARCHDIALOG_MATCH_CASE = auto()
    SEARCHDIALOG_WHOLE_WORD = auto()
    SEARCHDIALOG_WRAP = auto()
    SEARCHDIALOG_REGEX = auto()
    SEARCHDIALOG_MULTI_REPLACE = auto()
    WFDIALOG_SUSPECTS_ONLY = auto()
    WFDIALOG_IGNORE_CASE = auto()
    WFDIALOG_DISPLAY_TYPE = auto()
    WFDIALOG_SORT_TYPE = auto()
    WFDIALOG_ITALIC_THRESHOLD = auto()
    WFDIALOG_REGEX = auto()
    CHECKERDIALOG_SORT_TYPE_DICT = auto()
    CHECKERDIALOG_SUSPECTS_ONLY_DICT = auto()
    DIALOG_GEOMETRY = auto()
    ROOT_GEOMETRY = auto()
    ROOT_GEOMETRY_STATE = auto()
    DEFAULT_LANGUAGES = auto()
    JEEBIES_PARANOIA_LEVEL = auto()
    LEVENSHTEIN_DISTANCE = auto()
    FOOTNOTE_INDEX_STYLE = auto()
    WRAP_LEFT_MARGIN = auto()
    WRAP_RIGHT_MARGIN = auto()
    WRAP_BLOCKQUOTE_INDENT = auto()
    WRAP_BLOCKQUOTE_RIGHT_MARGIN = auto()
    WRAP_BLOCK_INDENT = auto()
    WRAP_POETRY_INDENT = auto()
    WRAP_INDEX_MAIN_MARGIN = auto()
    WRAP_INDEX_WRAP_MARGIN = auto()
    WRAP_INDEX_RIGHT_MARGIN = auto()
    TEXT_MARKUP_ITALIC = auto()
    TEXT_MARKUP_BOLD = auto()
    TEXT_MARKUP_SMALLCAPS = auto()
    TEXT_MARKUP_GESPERRT = auto()
    TEXT_MARKUP_FONT = auto()
    PAGESEP_AUTO_TYPE = auto()
    THEME_NAME = auto()
    TEAROFF_MENUS = auto()
    COMPOSE_HISTORY = auto()
    TEXT_FONT_FAMILY = auto()
    TEXT_FONT_SIZE = auto()
    SPELL_THRESHOLD = auto()
    UNMATCHED_NESTABLE = auto()
    UNICODE_BLOCK = auto()
    UNICODE_SEARCH_HISTORY = auto()
    SPLIT_TEXT_WINDOW = auto()
    SPLIT_TEXT_SASH_COORD = auto()
    IMAGE_INVERT = auto()
    IMAGE_FLOAT_GEOMETRY = auto()
    IMAGE_DOCK_SASH_COORD = auto()
    IMAGE_SCALE_FACTOR = auto()
    IMAGE_VIEWER_ALERT = auto()
    SCANNOS_FILENAME = auto()
    SCANNOS_HISTORY = auto()
    HIGHLIGHT_QUOTBRAC = auto()
    COLUMN_NUMBERS = auto()
    HTML_ITALIC_MARKUP = auto()
    HTML_BOLD_MARKUP = auto()
    HTML_GESPERRT_MARKUP = auto()
    HTML_FONT_MARKUP = auto()
    HTML_UNDERLINE_MARKUP = auto()
    HTML_SHOW_PAGE_NUMBERS = auto()
    HTML_IMAGE_UNIT = auto()
    HTML_IMAGE_OVERRIDE_EPUB = auto()
    HTML_IMAGE_ALIGNMENT = auto()
    HIGHLIGHT_PROOFERCOMMENT = auto()


class Preferences:
    """Handle setting/getting/saving/loading/defaulting preferences.

    Call `add` to create/define each preference, giving its default value,
    and optionally a callback function, e.g. if loading or setting a pref
    requires a UI change or other side effect. Once UI is initially ready,
    call `run_callbacks` to deal with all required side effects from loading
    the GGprefs file.

    Load/Save preferences in temporary file when testing.

    Attributes:
        dict: dictionary of values for prefs.
        defaults: dictionary of values for defaults.
        callbacks: dictionary of callbacks - typically a function to
          deal with side effect of loading/setting the pref, which will be
          called after all prefs have been loaded and UI is ready. Also called
          each time pref is changed.
        persistent_vars: dictionary of Persistent variables (inherit from Tk.Variable),
          stored here so the variable can be set whenever the pref is set. The Persistent
          classes below handle the reverse, i.e. they set the pref whenever the variable
          is set.
        prefsdir: directory containing user prefs & data files
    """

    def __init__(self) -> None:
        """Initialize by loading from JSON file."""
        self.dict: dict[PrefKey, Any] = {}
        self.defaults: dict[PrefKey, Any] = {}
        self.callbacks: dict[PrefKey, Callable[[Any], None]] = {}
        self.persistent_vars: dict[PrefKey, tk.Variable] = {}
        if is_x11():
            self.prefsdir = os.path.join(os.path.expanduser("~"), ".ggpreferences")
        elif is_windows():
            self.prefsdir = os.path.join(
                shell.SHGetFolderPath(0, shellcon.CSIDL_PERSONAL, 0, 0), "GGprefs"
            )
        else:
            self.prefsdir = os.path.join(
                os.path.expanduser("~"), "Documents", "GGprefs"
            )
        # If testing, use a test prefs file so tests and normal running do not interact
        if called_from_test:
            prefs_name = "GGprefs_test.json"
        else:
            prefs_name = "GGprefs.json"
        self.prefsfile = os.path.join(self.prefsdir, prefs_name)

        self._remove_test_prefs_file()
        self.permanent = False

    def get(self, key: PrefKey) -> Any:
        """Get preference value using key.

        Args:
            key: Name of preference.

        Returns:
            Preferences value; default for ``key`` if no preference set;
            ``None`` if no default for ``key``.
        """
        return copy.deepcopy(self.dict.get(key, self.defaults.get(key)))

    def set(self, key: PrefKey, value: Any) -> None:
        """Set preference value and save to file if value has changed.

        If key has an associated callback, call it.

        Args:
            key: Name of preference.
            value: Value for preference.
        """
        if self.get(key) == value:
            return
        self.dict[key] = value
        self.save()
        if key in self.callbacks:
            self.callbacks[key](value)
        if key in self.persistent_vars:
            self.persistent_vars[key].set(value)

    def toggle(self, key: PrefKey) -> None:
        """Toggle the value of a boolean preference.

        Args:
            key: Name of preference.
        """
        self.set(key, not self.get(key))

    def __del__(self) -> None:
        """Remove any test prefs file when finished."""
        self._remove_test_prefs_file()

    def set_default(self, key: PrefKey, default: Any) -> None:
        """Set default preference value.

        Args:
            key: Name of preference.
            default: Default value for preference.
        """
        self.defaults[key] = default

    def get_default(self, key: PrefKey) -> Any:
        """Get default preference value.

        Args:
            key: Name of preference.

        Returns:
            Default value for preference; ``None`` (with assertion) if no default for ``key``
        """
        assert key in self.defaults  # Raise assertion if no default set
        try:
            return self.defaults[key]
        except KeyError:
            return None

    def set_callback(self, key: PrefKey, callback: Callable[[Any], None]) -> None:
        """Add a callback for preference. Callback will be run when all
          prefs are loaded and UI is ready.

        Args:
            key: Name of preference.
            callback: Function to call if side effect required, e.g. to
              update UI to reflect a setting.
        """
        self.callbacks[key] = callback

    def link_persistent_var(self, key: PrefKey, var: tk.Variable) -> None:
        """Link a persistent variable to a preference. Variable will be set
        whenever the preference gets set.

        Args:
            key: Name of preference.
            var: Persistent variable to link to preference.
        """
        self.persistent_vars[key] = var

    def keys(self) -> list[PrefKey]:
        """Return list of preferences keys.

        Also includes preferences that have not been set, so only exist
        as defaults."""
        return list(set(list(self.dict.keys()) + list(self.defaults.keys())))

    def set_permanent(self, permanent: bool) -> None:
        """Set whether prefs should be loaded from and saved to Prefs file
        for permanent storage.

        Args:
            permanent: True if Prefs file should be used.
        """
        self.permanent = permanent

    def save(self) -> None:
        """Save preferences dictionary to JSON file."""

        if not self.permanent:
            return

        # Create Prefs dir (including any parent dirs) if needed
        try:
            os.makedirs(self.prefsdir, exist_ok=True)
        except OSError:
            logger.error(f"Unable to create {self.prefsdir}")
            return

        try:
            with open(self.prefsfile, "w", encoding="utf-8") as fp:
                json.dump(self.dict, fp, indent=2, ensure_ascii=False)
        except OSError:
            logger.error(f"Unable to save preferences to {self.prefsfile}")

    def load(self) -> None:
        """Load dictionary from JSON file, and use PrefKeys
        to store values in preferences dictionary."""

        if not self.permanent:
            return

        self.dict = {}
        if loaded_dict := load_dict_from_json(self.prefsfile):
            for key, value in loaded_dict.items():
                try:
                    self.dict[PrefKey(key)] = value
                except ValueError:
                    logger.debug(f"'{key}' is not a valid PrefKey - ignored")

    def run_callbacks(self) -> None:
        """Run all defined callbacks, passing value as argument.

        Should be called after prefs are loaded and UI is ready"""
        for key, callback in self.callbacks.items():
            if callback is not None:
                callback(self.get(key))

    def _remove_test_prefs_file(self) -> None:
        """Remove temporary JSON file used for prefs during testing."""
        if called_from_test and os.path.exists(self.prefsfile):
            os.remove(self.prefsfile)


class PersistentBoolean(tk.BooleanVar):
    """Tk boolean variable whose value is stored in user prefs file.

    Note that, like all prefs, the default value must be set in
    `initialize_preferences`
    """

    def __init__(self, prefs_key: PrefKey) -> None:
        """Initialize persistent boolean.

        Args:
            prefs_key: Preferences key associated with the variable.
        """
        super().__init__(value=preferences.get(prefs_key))
        self.trace_add("write", lambda *_args: preferences.set(prefs_key, self.get()))
        preferences.link_persistent_var(prefs_key, self)


class PersistentInt(tk.IntVar):
    """Tk integer variable whose value is stored in user prefs file.

    Note that, like all prefs, the default value must be set in
    `initialize_preferences`
    """

    def __init__(self, prefs_key: PrefKey) -> None:
        """Initialize persistent boolean.

        Args:
            prefs_key: Preferences key associated with the variable.
        """
        super().__init__(value=preferences.get(prefs_key))

        def set_pref(*_args: Any) -> None:
            """Set the preference if the IntVar holds an int.

            If it's not, e.g. because the entry field it is linked to is empty,
            or contains non-digits, a tk exception occurs, and we don't set the pref.
            """
            try:
                preferences.set(prefs_key, self.get())
            except tk.TclError:
                pass

        self.trace_add("write", set_pref)
        preferences.link_persistent_var(prefs_key, self)


class PersistentString(tk.StringVar):
    """Tk string variable whose value is stored in user prefs file.

    Note that, like all prefs, the default value must be set in
    `initialize_preferences`
    """

    def __init__(self, prefs_key: PrefKey) -> None:
        """Initialize persistent string.

        Args:
            prefs_key: Preferences key associated with the variable.
        """
        super().__init__(value=preferences.get(prefs_key))
        self.trace_add("write", lambda *_args: preferences.set(prefs_key, self.get()))
        preferences.link_persistent_var(prefs_key, self)


preferences = Preferences()



"""Store, analyze and report on word frequency and inconsistencies."""

from enum import StrEnum, auto
import tkinter as tk
from tkinter import ttk
from typing import Any, Callable

import regex as re

from guiguts.file import PAGE_SEPARATOR_REGEX
from guiguts.maintext import maintext
from guiguts.mainwindow import ScrolledReadOnlyText
from guiguts.misc_tools import tool_save
from guiguts.preferences import (
    preferences,
    PersistentBoolean,
    PersistentString,
    PrefKey,
)
from guiguts.search import SearchDialog
from guiguts.utilities import (
    sing_plur,
    DiacriticRemover,
    IndexRange,
    IndexRowCol,
    sound_bell,
    process_accel,
    cmd_ctrl_string,
    is_mac,
)
from guiguts.widgets import (
    ToplevelDialog,
    Combobox,
    ToolTip,
    mouse_bind,
    Busy,
)

_the_word_lists = None  # pylint: disable=invalid-name

RETURN_ARROW = "⏎"
MARKUP_TYPES = "i|b|sc|f|g|u|cite|em|strong"


class WFDisplayType(StrEnum):
    """Enum class to store Word Frequency display types."""

    ALL_WORDS = auto()
    EMDASHES = auto()
    HYPHENS = auto()
    ALPHANUM = auto()
    ALL_CAPS = auto()
    MIXED_CASE = auto()
    INITIAL_CAPS = auto()
    ACCENTS = auto()
    LIGATURES = auto()
    MARKEDUP = auto()
    CHAR_COUNTS = auto()
    REGEXP = auto()


class WFSortType(StrEnum):
    """Enum class to store Word Frequency sort types."""

    ALPHABETIC = auto()
    FREQUENCY = auto()
    LENGTH = auto()


class WFDict(dict[str, int]):
    """Dictionary containing word and frequency."""


class WFWordLists:
    """Word lists used for Word Frequency analysis.

    Attributes:
        all_words: All the words in the file.
        emdash_words: All pairs of words separated by an emdash/double hyphen.
    """

    def __init__(self) -> None:
        """Initialize a WFWordLists object to hold word frequency information."""
        self.reset()

    def reset(self) -> None:
        """Reset the word lists."""
        self.all_words: WFDict = WFDict()
        self.emdash_words: WFDict = WFDict()

    def ensure_file_analyzed(self) -> None:
        """Analyze file to create word lists, unless already done.

        Call reset method first to force reanalysis."""
        if self.all_words:
            return
        for line, _ in maintext().get_lines():
            if re.search(PAGE_SEPARATOR_REGEX, line):
                continue
            if preferences.get(PrefKey.WFDIALOG_IGNORE_CASE):
                line = line.lower()
            line = re.sub(r"<\/?[a-z]*>", " ", line)  # throw away DP tags
            # get rid of nonalphanumeric (retaining combining characters)
            line = re.sub(r"[^'’\.,\p{Alnum}\p{Mark}*_-]", " ", line)

            def strip_punc(word: str) -> str:
                """Strip relevant leading/trailing punctuation from word."""
                return re.sub(r"^[\.,'’_-]+|[\.,'’_-]+$", "", word)

            # Build a list of emdash words, i.e. "word1--word2"
            words = re.split(r"\s+", line)
            for word in words:
                if re.search(r"[^-](--|—)[^-]", word) and "---" not in word:
                    word = strip_punc(word)
                    tally_word(self.emdash_words, word)

            line = re.sub(r"(--|—)", " ", line)  # double-hyphen/emdash

            words = re.split(r"\s+", line)
            for word in words:
                word = re.sub(r"(?<!\-)\*", "", word)  # * not preceded by hyphen
                word = re.sub(r"^\*$", "", word)  # just *
                word = strip_punc(word)
                # Tally single word
                tally_word(self.all_words, word)

    def get_all_words(self) -> WFDict:
        """Return the list of all words in the file.

        Returns:
            Dictionary of words with their frequencies."""
        self.ensure_file_analyzed()
        return self.all_words

    def get_emdash_words(self) -> WFDict:
        """Return the list of emdash words (word1--word2) in the file.

        Returns:
            Dictionary of emdash words with their frequencies."""
        self.ensure_file_analyzed()
        return self.emdash_words


class WordFrequencyEntry:
    """Class to hold one entry in the Word Frequency dialog.

    Attributes:
        text: Single line of text to display in dialog.
        count: Number of occurrences.
        suspect: True if entry is a "suspect".
    """

    SUSPECT = "****"

    def __init__(self, word: str, frequency: int, suspect: bool) -> None:
        """Initialize WordFrequencyEntry object.

        Args:
            word: Word for this entry.
            frequency: Number of occurrences of word.
        """
        self.word = word
        self.frequency = frequency
        self.suspect = suspect


class WordFrequencyDialog(ToplevelDialog):
    """Dialog to show results of word frequency analysis.

    Attributes:
        text: Text widget to contain results.
    """

    manual_page = "Tools_Menu#Word_Frequency"
    CHAR_DISPLAY: dict[str, str] = {
        " ": "*space*",
        "\u00a0": "*nbsp*",
        "\n": RETURN_ARROW,
    }

    def __init__(
        self,
    ) -> None:
        """Initialize the dialog."""
        super().__init__("Word Frequency")
        self.top_frame.rowconfigure(0, weight=0)
        header_frame = ttk.Frame(self.top_frame, padding=2)
        header_frame.grid(row=0, column=0, sticky="NSEW")
        header_frame.columnconfigure(0, weight=1)

        # Message label
        label_frame = ttk.Frame(header_frame, borderwidth=1, relief=tk.GROOVE)
        label_frame.grid(row=0, column=0, sticky="NSEW")
        # label_frame.columnconfigure(0, weight=1)
        self.message = tk.StringVar()
        ttk.Label(label_frame, textvariable=self.message).grid(
            row=0, column=0, sticky="NSW", padx=5, pady=2
        )

        # Re-run buttons
        rerun_frame = ttk.Frame(header_frame, borderwidth=1, relief=tk.GROOVE)
        rerun_frame.grid(row=0, column=1, rowspan=2, padx=(0, 15))
        ttk.Button(
            rerun_frame, text="Re-run", command=word_frequency, takefocus=False
        ).grid(row=0, column=0, sticky="NSEW", padx=5, pady=2)

        def change_ignore_case() -> None:
            """Handle changing of ignore case flag - enable/disable buttons & re-run tool."""
            self.set_case_sensitive_btns()
            word_frequency()

        ttk.Checkbutton(
            rerun_frame,
            text="Ignore Case",
            command=change_ignore_case,
            variable=PersistentBoolean(PrefKey.WFDIALOG_IGNORE_CASE),
            takefocus=False,
        ).grid(row=1, column=0, sticky="NSEW", padx=5, pady=2)

        # Options
        options_frame = ttk.Frame(
            header_frame, borderwidth=1, relief=tk.GROOVE, padding=2
        )
        options_frame.grid(row=1, column=0, sticky="NSEW")
        options_frame.columnconfigure(0, weight=1)
        self.suspects_btn = ttk.Checkbutton(
            options_frame,
            text="Suspects Only",
            variable=PersistentBoolean(PrefKey.WFDIALOG_SUSPECTS_ONLY),
            command=lambda: wf_populate(self),
            takefocus=False,
        )
        self.suspects_btn.grid(row=0, column=0, sticky="NSW", padx=5)

        def copy_errors() -> None:
            """Copy text messages to clipboard."""
            self.clipboard_clear()
            self.clipboard_append(self.text.get("1.0", tk.END))

        copy_button = ttk.Button(
            options_frame, text="Copy Results", command=copy_errors
        )
        copy_button.grid(row=0, column=1, sticky="NSE", padx=(0, 20))

        ttk.Label(
            options_frame,
            text="Sort:",
        ).grid(row=0, column=2, sticky="NSE", padx=5)
        sort_type = PersistentString(PrefKey.WFDIALOG_SORT_TYPE)
        ttk.Radiobutton(
            options_frame,
            text="Alph",
            command=lambda: wf_populate(self),
            variable=sort_type,
            value=WFSortType.ALPHABETIC,
            takefocus=False,
        ).grid(row=0, column=3, sticky="NSE", padx=2)
        ttk.Radiobutton(
            options_frame,
            text="Freq",
            command=lambda: wf_populate(self),
            variable=sort_type,
            value=WFSortType.FREQUENCY,
            takefocus=False,
        ).grid(row=0, column=4, sticky="NSE", padx=2)
        ttk.Radiobutton(
            options_frame,
            text="Len",
            command=lambda: wf_populate(self),
            variable=sort_type,
            value=WFSortType.LENGTH,
            takefocus=False,
        ).grid(row=0, column=5, sticky="NSE", padx=(2, 5))

        # Display type radio buttons
        display_frame = ttk.Frame(
            header_frame, borderwidth=1, relief=tk.GROOVE, padding=2
        )
        display_frame.grid(row=2, column=0, columnspan=2, sticky="NSEW", padx=(0, 15))
        for col in range(0, 4):
            display_frame.columnconfigure(index=col, weight=1)

        display_type = PersistentString(PrefKey.WFDIALOG_DISPLAY_TYPE)

        def display_radio(
            row: int,
            column: int,
            text: str,
            value: str,
            frame: ttk.Frame = display_frame,
        ) -> ttk.Radiobutton:
            """Add a radio button to change display type.

            Args:
                row: Row to place button.
                column: Column to place button.
                text: Text for button.
                value: Value to be set when button is selected - a constant from WFDisplayType
            """
            button = ttk.Radiobutton(
                frame,
                text=text,
                command=lambda: wf_populate(self),
                variable=display_type,
                value=value,
                takefocus=False,
            )
            button.grid(row=row, column=column, sticky="NSW", padx=5)
            return button

        display_radio(0, 0, "All Words", WFDisplayType.ALL_WORDS)
        display_radio(0, 1, "Emdashes", WFDisplayType.EMDASHES)
        display_radio(0, 2, "Hyphens", WFDisplayType.HYPHENS)
        self.all_caps_btn = display_radio(1, 0, "ALL CAPITALS", WFDisplayType.ALL_CAPS)
        self.mixed_case_btn = display_radio(
            1, 1, "MiXeD CasE", WFDisplayType.MIXED_CASE
        )
        self.initial_caps_btn = display_radio(
            1, 2, "Initial Capitals", WFDisplayType.INITIAL_CAPS
        )
        display_radio(2, 0, "Alpha/Num", WFDisplayType.ALPHANUM)
        display_radio(2, 1, "Accents", WFDisplayType.ACCENTS)
        display_radio(2, 2, "Ligatures", WFDisplayType.LIGATURES)
        display_radio(3, 0, "Character Cnts", WFDisplayType.CHAR_COUNTS)
        italic_frame = ttk.Frame(display_frame)
        italic_frame.grid(row=3, column=1, columnspan=2, sticky="NSEW")
        italic_frame.columnconfigure(index=1, weight=1)
        display_radio(
            0,
            0,
            "Ital/Bold/SC/etc - Word Threshold:",
            WFDisplayType.MARKEDUP,
            italic_frame,
        )

        def is_nonnegative_int(new_value: str) -> bool:
            """Validation routine for Combobox - check value is a non-negative integer,
            also allowing empty string."""
            return re.fullmatch(r"\d*", new_value) is not None

        self.threshold_box = Combobox(
            italic_frame,
            PrefKey.WFDIALOG_ITALIC_THRESHOLD,
            width=6,
            validate=tk.ALL,
            validatecommand=(self.register(is_nonnegative_int), "%P"),
        )
        self.threshold_box.grid(row=0, column=1, sticky="NSEW", padx=(0, 5))
        self.threshold_box.display_latest_value()
        ToolTip(
            self.threshold_box,
            "Only show marked up phrases with no more than this number of words",
        )
        # Override default behavior for combobox up/down arrow
        self.threshold_box.bind("<Up>", lambda _e: self.goto_word_by_arrow(-1))
        self.threshold_box.bind("<Down>", lambda _e: self.goto_word_by_arrow(1))

        def display_markedup(*_args: Any) -> None:
            """Callback to display the marked up words with new threshold value."""
            preferences.set(PrefKey.WFDIALOG_DISPLAY_TYPE, WFDisplayType.MARKEDUP)
            wf_populate(self)

        self.threshold_box.bind("<Return>", display_markedup)
        self.threshold_box.bind("<<ComboboxSelected>>", display_markedup)

        display_radio(4, 0, "Regular Expression", WFDisplayType.REGEXP)
        self.regex_box = Combobox(display_frame, PrefKey.WFDIALOG_REGEX)
        self.regex_box.grid(row=4, column=1, columnspan=2, sticky="NSEW", padx=(0, 5))
        ToolTip(self.regex_box, "Only show words matching this regular expression")
        # Override default behavior for combobox up/down arrow
        self.regex_box.bind("<Up>", lambda _e: self.goto_word_by_arrow(-1))
        self.regex_box.bind("<Down>", lambda _e: self.goto_word_by_arrow(1))

        def display_regexp(*_args: Any) -> None:
            """Callback to display the regex-matching words with new regex."""
            preferences.set(PrefKey.WFDIALOG_DISPLAY_TYPE, WFDisplayType.REGEXP)
            wf_populate(self)

        self.regex_box.bind("<Return>", display_regexp)
        self.regex_box.bind("<<ComboboxSelected>>", display_regexp)
        self.regex_box.display_latest_value()

        # Main display list
        self.top_frame.rowconfigure(1, weight=1)
        self.text = ScrolledReadOnlyText(
            self.top_frame,
            context_menu=False,
            wrap=tk.NONE,
            font=maintext().font,
        )
        self.text.grid(row=1, column=0, sticky="NSEW")
        mouse_bind(self.text, "1", self.goto_word_by_click)
        _, event = process_accel("Cmd/Ctrl+1")
        self.text.bind(event, self.search_word)
        self.text.bind("<Key>", self.goto_word_by_letter)
        _, event = process_accel("Cmd/Ctrl+a")
        self.text.bind(event, lambda _e: self.text.event_generate("<<SelectAll>>"))
        _, event = process_accel("Cmd/Ctrl+A")
        self.text.bind(event, lambda _e: self.text.event_generate("<<SelectAll>>"))
        _, event = process_accel("Cmd/Ctrl+c")
        self.text.bind(event, lambda _e: self.text.event_generate("<<Copy>>"))
        _, event = process_accel("Cmd/Ctrl+C")
        self.text.bind(event, lambda _e: self.text.event_generate("<<Copy>>"))
        self.text.bind("<Home>", lambda _e: self.goto_word(0))
        self.text.bind("<Shift-Home>", lambda _e: self.goto_word(0))
        self.text.bind("<End>", lambda _e: self.goto_word(len(self.entries) - 1))
        self.text.bind("<Shift-End>", lambda _e: self.goto_word(len(self.entries) - 1))
        # Bind same keys as main window uses for top/bottom on Mac.
        # Above bindings work already for Windows
        if is_mac():
            self.text.bind("<Command-Up>", lambda _e: self.goto_word(0))
            self.text.bind("<Shift-Command-Up>", lambda _e: self.goto_word(0))
            self.text.bind(
                "<Command-Down>", lambda _e: self.goto_word(len(self.entries) - 1)
            )
            self.text.bind(
                "<Shift-Command-Down>", lambda _e: self.goto_word(len(self.entries) - 1)
            )
        ToolTip(
            self.text,
            "\n".join(
                [
                    "Left click: Find first match, click again for other matches",
                    f"{cmd_ctrl_string()}-left click: Find using Search dialog",
                ]
            ),
            use_pointer_pos=True,
        )

        self.bind("<Up>", lambda _e: self.goto_word_by_arrow(-1))
        self.bind("<Down>", lambda _e: self.goto_word_by_arrow(1))

        self.previous_word = ""
        # Store tooltips so they can be added/destroyed depending on Ignore Case
        self.tooltip_dict: dict[ttk.Widget, ToolTip] = {}
        self.set_case_sensitive_btns()

        self.minsize(450, 100)
        self.reset()

    def reset(self) -> None:
        """Reset dialog."""
        super().reset()
        self.entries: list[WordFrequencyEntry] = []
        if maintext().winfo_exists():
            maintext().remove_spotlights()
        if not self.text.winfo_exists():
            return
        self.text.delete("1.0", tk.END)
        self.message.set("")

    def set_case_sensitive_btns(self) -> None:
        """Enable/disable buttons depending on Ignore Case setting.

        Show/hide a tooltip to explain to the user.
        """
        ignore_case = preferences.get(PrefKey.WFDIALOG_IGNORE_CASE)
        for widget in (self.all_caps_btn, self.mixed_case_btn, self.initial_caps_btn):
            widget["state"] = tk.DISABLED if ignore_case else tk.NORMAL
            if ignore_case:
                self.tooltip_dict[widget] = ToolTip(
                    widget, "Cannot be used with 'Ignore Case'"
                )
            else:
                try:
                    self.tooltip_dict[widget].destroy()
                except (tk.TclError, KeyError):
                    pass  # OK for tooltip not to exist

    def add_entry(self, word: str, frequency: int, suspect: bool = False) -> None:
        """Add an entry to be displayed in the dialog.

        Args:
            word: Word for this entry.
            frequency: Number of occurrences of word.
            suspect: Optional bool to flag this word as "suspect".
        """
        entry = WordFrequencyEntry(word, frequency, suspect)
        self.entries.append(entry)

    def display_entries(self) -> None:
        """Display all the stored entries in the dialog according to
        the sort setting."""

        def sort_key_alpha(
            entry: WordFrequencyEntry,
        ) -> tuple[str, ...]:
            no_dia = DiacriticRemover.remove_diacritics(entry.word)
            return (no_dia.lower(), no_dia, entry.word)

        def sort_key_alpha_no_markup(
            entry: WordFrequencyEntry,
        ) -> tuple[str, ...]:
            unmarked = re.sub(rf"^<({MARKUP_TYPES})>", "", entry.word)
            unmarked = re.sub(rf"</({MARKUP_TYPES})>$", "", unmarked)
            unmarked_no_dia = DiacriticRemover.remove_diacritics(unmarked)
            no_dia = DiacriticRemover.remove_diacritics(entry.word)
            return (
                unmarked_no_dia.lower(),
                unmarked_no_dia,
                no_dia.lower(),
                no_dia,
                entry.word,
            )

        def sort_key_freq(entry: WordFrequencyEntry) -> tuple[int | str, ...]:
            no_dia = DiacriticRemover.remove_diacritics(entry.word)
            return (-entry.frequency,) + (no_dia.lower(), no_dia, entry.word)

        def sort_key_len(entry: WordFrequencyEntry) -> tuple[int | str, ...]:
            no_dia = DiacriticRemover.remove_diacritics(entry.word)
            return (-len(entry.word), no_dia.lower(), no_dia, entry.word)

        key: Callable[[WordFrequencyEntry], tuple]
        match preferences.get(PrefKey.WFDIALOG_SORT_TYPE):
            case WFSortType.ALPHABETIC:
                key = (
                    sort_key_alpha_no_markup
                    if preferences.get(PrefKey.WFDIALOG_DISPLAY_TYPE)
                    == WFDisplayType.MARKEDUP
                    else sort_key_alpha
                )
            case WFSortType.FREQUENCY:
                key = sort_key_freq
            case WFSortType.LENGTH:
                key = sort_key_len
            case _ as bad_value:
                assert False, f"Invalid WFSortType: {bad_value}"

        # Sort stored list, rather than just displayed list, since later
        # we'll want to index into list based on index in display.
        self.entries.sort(key=key)
        # Get longest frequency to aid formatting
        max_freq = 0
        for entry in self.entries:
            max_freq = max(max_freq, entry.frequency)
        max_freq_len = len(str(max_freq))
        # Display entries
        for entry in self.entries:
            suspect = f" {WordFrequencyEntry.SUSPECT}" if entry.suspect else ""
            # Single whitespace characters are replaced with a visible label
            try:
                word = WordFrequencyDialog.CHAR_DISPLAY[entry.word]
            except KeyError:
                word = entry.word
            message = f"{entry.frequency:>{max_freq_len}}  {word}{suspect}\n"
            self.text.insert(tk.END, message)

    def whole_word_search(self, word: str) -> bool:
        """Return if a whole word search should be done for given word.

        Args:
            word: The "word" being searched for (may be multiple word string).

        Returns:
            True if "word" should be searched for with "wholeword" flag. This is
            not the case for character count or words that begin/end with
            a non-word character."""
        return not (
            preferences.get(PrefKey.WFDIALOG_DISPLAY_TYPE) == WFDisplayType.CHAR_COUNTS
            or re.search(r"^\W|\W$", word)
        )

    def goto_word_by_click(self, event: tk.Event) -> str:
        """Go to first/next occurrence of word selected by `event`.

        If a different word to last click, then start search from beginning
        of file, otherwise just continue from current location.

        Args:
            event: Event containing mouse click coordinates.

        Returns:
            "break" to stop further processing of event
        """
        try:
            entry_index = self.entry_index_from_click(event)
        except IndexError:
            return "break"
        self.goto_word(entry_index)
        return "break"

    def goto_word_by_letter(self, event: tk.Event) -> str:
        """Go to first occurrence of word beginning with the character in `event`.

        Args:
            event: Event containing keystroke.

        Returns:
            "break" to stop further processing of events if valid character
        """
        if not event.char:
            return ""
        low_char = event.char.lower()
        for idx, entry in enumerate(self.entries):
            if entry.word.lower().startswith(low_char):
                self.text.select_line(idx + 1)
                return "break"
        return ""

    def goto_word_by_arrow(self, increment: int) -> str:
        """Select next/previous line in dialog, and jump to the line in the
        main text widget that corresponds to it.

        Args:
            increment: +1 to move to next line, -1 to move to previous line.
        """
        entry_index = self.text.get_select_line_num()
        if entry_index is not None:
            entry_index -= 1
        if (
            entry_index is None
            or entry_index + increment < 0
            or entry_index + increment >= len(self.entries)
        ):
            return "break"
        self.goto_word(entry_index + increment)
        return "break"

    def goto_word(self, entry_index: int, force_first: bool = False) -> None:
        """Go to first/next occurrence of word listed at `entry_index`.

        If a different word to last click, then start search from beginning
        of file, otherwise just continue from current location.

        Args:
            entry_index: Index into `self.entries` indicating which to select.
            force_first: True to force first occurrence, False for next occurrence.
        """
        if entry_index >= len(self.entries):
            return
        self.text.tag_remove("sel", "1.0", tk.END)
        self.text.select_line(entry_index + 1)
        self.text.mark_set(tk.INSERT, f"{entry_index + 1}.0")
        self.text.focus()
        word = self.entries[entry_index].word
        if word == self.previous_word and not force_first:
            start = maintext().get_insert_index()
            start.col += 1
        else:
            start = maintext().start()
        # Special handling for newline characters (displayed as RETURN_ARROW)
        newline_word = (
            re.sub(RETURN_ARROW, "\n", word) if RETURN_ARROW in word else word
        )
        # Also special handling for non-marked-up phrase from marked-up check
        # (Regex to specifically find non-marked-up version), for char count,
        # and for words that begin/end with non-word chars
        if preferences.get(
            PrefKey.WFDIALOG_DISPLAY_TYPE
        ) == WFDisplayType.MARKEDUP and not word.startswith("<"):
            regexp = True
            match_word = r"(^|[^>\w])" + re.escape(newline_word) + r"($|[^<\w])"
        elif (
            preferences.get(PrefKey.WFDIALOG_DISPLAY_TYPE) == WFDisplayType.CHAR_COUNTS
        ):
            regexp = False
            match_word = newline_word
        elif self.whole_word_search(word):
            regexp = True
            match_word = (
                rf"\y{re.escape(newline_word)}\y"  # Tcl regex: \y is word boundary
            )
        else:  # Word begins/ends with non-word char - do manual whole-word
            regexp = True
            match_word = r"(^|\W)" + re.escape(newline_word) + r"($|\W)"

        # If hyphen matching and there's one (escaped) space in the word, let that match
        # any amount of whitespace
        if (
            preferences.get(PrefKey.WFDIALOG_DISPLAY_TYPE) == WFDisplayType.HYPHENS
            and r"\ " in match_word
        ):
            regexp = True
            match_word = match_word.replace(r"\ ", r"[\n ]+")

        match = maintext().find_match(
            match_word,
            IndexRange(start, maintext().end()),
            nocase=preferences.get(PrefKey.WFDIALOG_IGNORE_CASE),
            regexp=regexp,
            backwards=False,
        )
        if match is None:
            sound_bell()
        else:
            # If searched with regex, it may have matched up to 1 extra character each side,
            # so adjust match length (count) and start column to match the actual word.
            if regexp:
                match_str = maintext().get_match_text(match)
                match.count = len(newline_word)
                if preferences.get(PrefKey.WFDIALOG_IGNORE_CASE):
                    if match_str[1 : match.count + 1].lower() == newline_word.lower():
                        match.rowcol.col += 1
                else:
                    if match_str[1 : match.count + 1] == newline_word:
                        match.rowcol.col += 1
            maintext().set_insert_index(match.rowcol, focus=False)
            maintext().remove_spotlights()
            start_index = match.rowcol.index()
            end_index = maintext().index(start_index + f"+{match.count}c")
            maintext().spotlight_range(IndexRange(start_index, end_index))
            self.previous_word = word

    def search_word(self, event: tk.Event) -> str:
        """Open search dialog ready to search for selected word."""
        try:
            entry_index = self.entry_index_from_click(event)
        except IndexError:
            return "break"
        self.text.select_line(entry_index + 1)
        word = self.entries[entry_index].word

        # Special handling for newline characters (displayed as RETURN_ARROW)
        match_word = re.sub(RETURN_ARROW, "\n", word) if RETURN_ARROW in word else word

        preferences.set(
            PrefKey.SEARCHDIALOG_MATCH_CASE,
            not preferences.get(PrefKey.WFDIALOG_IGNORE_CASE),
        )
        preferences.set(PrefKey.SEARCHDIALOG_WHOLE_WORD, self.whole_word_search(word))
        preferences.set(PrefKey.SEARCHDIALOG_REGEX, False)
        dlg = SearchDialog.show_dialog()
        dlg.search_box_set(match_word)
        dlg.display_message()

        return "break"

    def entry_index_from_click(self, event: tk.Event) -> int:
        """Get the index into the list of entries based on the mouse position
        in the click event.

        Args:
            event: Event object containing mouse click position.

        Returns:
            Index into self.entries list
            Raises IndexError exception if out of range
        """
        entry_index = IndexRowCol(self.text.index(f"@{event.x},{event.y}")).row - 1
        if entry_index < 0 or entry_index >= len(self.entries):
            raise IndexError
        return entry_index


def word_frequency() -> None:
    """Do word frequency analysis on file."""
    global _the_word_lists

    if not tool_save():
        return

    _the_word_lists = WFWordLists()

    wf_dialog = WordFrequencyDialog.show_dialog()
    wf_populate(wf_dialog)


def wf_populate(wf_dialog: WordFrequencyDialog) -> None:
    """Populate the WF dialog with words based on the display type.

    Args:
        wf_dialog: The word frequency dialog.
    """
    Busy.busy()
    wf_dialog.previous_word = ""
    display_type = preferences.get(PrefKey.WFDIALOG_DISPLAY_TYPE)

    # Suspects Only is only relevant for some modes
    if display_type in (
        WFDisplayType.EMDASHES,
        WFDisplayType.HYPHENS,
        WFDisplayType.MARKEDUP,
        WFDisplayType.ACCENTS,
        WFDisplayType.LIGATURES,
    ):
        wf_dialog.suspects_btn.grid()
    else:
        wf_dialog.suspects_btn.grid_remove()

    match display_type:
        case WFDisplayType.ALL_WORDS:
            wf_populate_all(wf_dialog)
        case WFDisplayType.EMDASHES:
            wf_populate_emdashes(wf_dialog)
        case WFDisplayType.HYPHENS:
            wf_populate_hyphens(wf_dialog)
        case WFDisplayType.ALPHANUM:
            wf_populate_alphanum(wf_dialog)
        case WFDisplayType.ALL_CAPS:
            wf_populate_allcaps(wf_dialog)
        case WFDisplayType.MIXED_CASE:
            wf_populate_mixedcase(wf_dialog)
        case WFDisplayType.INITIAL_CAPS:
            wf_populate_initialcaps(wf_dialog)
        case WFDisplayType.MARKEDUP:
            wf_populate_markedup(wf_dialog)
        case WFDisplayType.ACCENTS:
            wf_populate_accents(wf_dialog)
        case WFDisplayType.LIGATURES:
            wf_populate_ligatures(wf_dialog)
        case WFDisplayType.CHAR_COUNTS:
            wf_populate_charcounts(wf_dialog)
        case WFDisplayType.REGEXP:
            wf_populate_regexps(wf_dialog)
        case _ as bad_value:
            assert False, f"Invalid WFDisplayType: {bad_value}"
    wf_dialog.goto_word(0)
    Busy.unbusy()


def wf_populate_all(wf_dialog: WordFrequencyDialog) -> None:
    """Populate the WF dialog with the list of all words.

    Args:
        wf_dialog: The word frequency dialog.
    """
    assert _the_word_lists is not None
    wf_dialog.reset()

    all_words = _the_word_lists.get_all_words()
    total_cnt = 0
    for word, freq in all_words.items():
        wf_dialog.add_entry(word, freq)
        total_cnt += freq
    wf_dialog.display_entries()
    wf_dialog.message.set(
        f"{sing_plur(total_cnt, 'total word')}; {sing_plur(len(all_words), 'distinct word')}"
    )


def wf_populate_emdashes(wf_dialog: WordFrequencyDialog) -> None:
    """Populate the WF dialog with the list of emdashed words.

    Args:
        wf_dialog: The word frequency dialog.
    """
    assert _the_word_lists is not None
    wf_dialog.reset()

    all_words = _the_word_lists.get_all_words()
    emdash_words = _the_word_lists.get_emdash_words()
    suspect_cnt = 0
    for emdash_word, freq in emdash_words.items():
        # Check for suspect, i.e. also seen with a single hyphen
        word = re.sub("(--|—)", "-", emdash_word)
        word_output = False
        if not preferences.get(PrefKey.WFDIALOG_SUSPECTS_ONLY):
            wf_dialog.add_entry(emdash_word, freq)
            word_output = True
        if word in all_words:
            if not word_output:
                wf_dialog.add_entry(emdash_word, freq)
            wf_dialog.add_entry(word, all_words[word], suspect=True)
            suspect_cnt += 1
    wf_dialog.display_entries()
    wf_dialog.message.set(
        f"{sing_plur(len(emdash_words), 'emdash phrase')}; {sing_plur(suspect_cnt, 'suspect')} ({WordFrequencyEntry.SUSPECT})"
    )


def wf_populate_hyphens(wf_dialog: WordFrequencyDialog) -> None:
    """Populate the WF dialog with the list of word pairs.

    Args:
        wf_dialog: The word frequency dialog.
    """
    assert _the_word_lists is not None
    wf_dialog.reset()

    all_words = _the_word_lists.get_all_words()
    emdash_words = _the_word_lists.get_emdash_words()

    # See if word pair suspects exist, e.g. "flash light" for "flash-light"
    word_pairs: WFDict = WFDict()
    # Replace single newline or multiple spaces with single space
    # (Multiple newlines is probably deliberate rather than error)
    whole_text = re.sub(r"(\n| +)", " ", maintext().get_text())
    for word in all_words:
        if "-" in word:
            pair = word.replace("-", " ")
            # Find word pair not preceded/followed by a letter - this is so
            # that space or punctuation surrounding the pair doesn't break things.
            count = len(re.findall(rf"(?<!\w){pair}(?!\w)", whole_text))
            if count:
                word_pairs[pair] = count

    suspect_cnt = 0
    total_cnt = 0
    word_output = {}
    for word, freq in all_words.items():
        if "-" not in word:
            continue
        total_cnt += 1
        # Check for suspects - given "w1-w2", then "w1w2", "w1 w2" and "w1--w2" are suspects.
        word_pair = re.sub(r"-\*?", " ", word)
        nohyp_word = re.sub(r"-\*?", "", word)
        twohyp_word = re.sub(r"-\*?", "--", word)
        suspect = (
            word_pair in word_pairs
            or nohyp_word in all_words
            or twohyp_word in emdash_words
        )

        if not preferences.get(PrefKey.WFDIALOG_SUSPECTS_ONLY):
            wf_dialog.add_entry(word, freq, suspect=suspect)
            word_output[word] = True
        if word_pair in word_pairs:
            if word not in word_output:
                wf_dialog.add_entry(word, freq, suspect=suspect)
                word_output[word] = True
            if word_pair not in word_output:
                wf_dialog.add_entry(word_pair, word_pairs[word_pair], suspect=suspect)
                word_output[word_pair] = True
                suspect_cnt += 1
        nohyp_word = re.sub(r"-\*?", "", word)
        if nohyp_word in all_words:
            if word not in word_output:
                wf_dialog.add_entry(word, freq, suspect=suspect)
                word_output[word] = True
            if nohyp_word not in word_output:
                wf_dialog.add_entry(nohyp_word, all_words[nohyp_word], suspect=suspect)
                word_output[nohyp_word] = True
                suspect_cnt += 1
        twohyp_word = re.sub(r"-\*?", "--", word)
        if twohyp_word in emdash_words:
            if word not in word_output:
                wf_dialog.add_entry(word, freq, suspect=suspect)
                word_output[word] = True
            if twohyp_word not in word_output:
                wf_dialog.add_entry(
                    twohyp_word, emdash_words[twohyp_word], suspect=suspect
                )
                word_output[twohyp_word] = True
                suspect_cnt += 1
    wf_dialog.display_entries()
    wf_dialog.message.set(
        f"{sing_plur(total_cnt, 'hyphenated word')}; {sing_plur(suspect_cnt, 'suspect')} ({WordFrequencyEntry.SUSPECT})"
    )


def wf_populate_by_match(
    wf_dialog: WordFrequencyDialog,
    desc: str,
    match_func: Callable[[str], Any],
) -> None:
    """Populate the WF dialog with the list of all words that match_func matches.

    Args:
        wf_dialog: The word frequency dialog.
        desc: Description for message, e.g. desc "ALLCAPS" -> message "27 ALLCAPS words"
        match_func: Function that returns Truthy result if word matches the required criteria
    """
    assert _the_word_lists is not None
    wf_dialog.reset()

    all_words = _the_word_lists.get_all_words()
    count = 0
    for word, freq in all_words.items():
        if match_func(word):
            wf_dialog.add_entry(word, freq)
            count += 1
    wf_dialog.display_entries()
    wf_dialog.message.set(f"{sing_plur(count, desc + ' word')}")


def wf_populate_alphanum(wf_dialog: WordFrequencyDialog) -> None:
    """Populate the WF dialog with the list of all alphanumeric words.

    Args:
        wf_dialog: The word frequency dialog.
    """
    wf_populate_by_match(
        wf_dialog,
        "alphanumeric",
        lambda word: re.search(r"\d", word) and re.search(r"\p{Alpha}", word),
    )


def wf_populate_allcaps(wf_dialog: WordFrequencyDialog) -> None:
    """Populate the WF dialog with the list of all ALLCAPS words.

    Args:
        wf_dialog: The word frequency dialog.
    """
    wf_populate_by_match(
        wf_dialog,
        "ALLCAPS",
        lambda word: re.search(r"\p{IsUpper}", word)
        and not re.search(r"\p{IsLower}", word),
    )


def wf_populate_mixedcase(wf_dialog: WordFrequencyDialog) -> None:
    """Populate the WF dialog with the list of all MiXeD CasE words.

    Args:
        wf_dialog: The word frequency dialog.
    """
    wf_populate_by_match(
        wf_dialog,
        "MiXeD CasE",
        lambda word: re.search(r"\p{IsUpper}", word)
        and re.search(r"\p{IsLower}", word)
        and not re.fullmatch(r"\p{Upper}[\p{IsLower}\p{Mark}\d'’*-]+", word),
    )


def wf_populate_initialcaps(wf_dialog: WordFrequencyDialog) -> None:
    """Populate the WF dialog with the list of all Initial Caps words.

    Args:
        wf_dialog: The word frequency dialog.
    """
    wf_populate_by_match(
        wf_dialog,
        "Initial Caps",
        lambda word: re.fullmatch(r"\p{Upper}\P{Upper}+", word),
    )


def wf_populate_markedup(wf_dialog: WordFrequencyDialog) -> None:
    """Populate the WF dialog with the list of all phrases marked up
    with DP markup, e.g. <i>, <b>, etc.

    Args:
        wf_dialog: The word frequency dialog.
    """
    wf_dialog.reset()

    marked_dict: WFDict = WFDict()
    nocase = preferences.get(PrefKey.WFDIALOG_IGNORE_CASE)
    search_flags = re.IGNORECASE if nocase else 0

    whole_text = maintext().get_text()

    matches = re.findall(
        rf"(?<!\w)(<({MARKUP_TYPES})>([^<]|\n)+</\2>)(?!\w)",
        whole_text,
        flags=search_flags,
    )
    for match in matches:
        marked_phrase: str = match[0]
        if nocase:
            marked_phrase = marked_phrase.lower()
        marked_phrase = marked_phrase.replace("\n", RETURN_ARROW)
        tally_word(marked_dict, marked_phrase)

    total_cnt = 0
    suspect_cnt = 0
    unmarked_count = {}
    for marked_phrase, marked_count in marked_dict.items():
        # Suspect if the bare phrase appears an excess number of times,
        # i.e. all occurrences > marked up occurrences
        unmarked_phrase = re.sub(rf"^<({MARKUP_TYPES})>", "", marked_phrase)
        unmarked_phrase = re.sub(rf"</({MARKUP_TYPES})>$", "", unmarked_phrase)
        unmarked_search = unmarked_phrase.replace(RETURN_ARROW, "\n")
        unmarked_search = r"(^|[^>\w])" + re.escape(unmarked_search) + r"($|[^<\w])"
        num_words = len(unmarked_search.split())
        # If phrase is longer than "threshold" words, skip it - zero/empty threshold allows any length
        threshold_str = wf_dialog.threshold_box.get()
        wf_dialog.threshold_box.add_to_history(threshold_str)
        threshold = int(threshold_str) if threshold_str else 0
        if num_words > threshold > 0:
            continue
        total_cnt += 1

        # Store unmarked counts so we don't do unmarked check twice,
        # e.g. if <i>dog</i>, <b>dog</b> and dog all exist
        if unmarked_phrase not in unmarked_count:
            unmarked_count[unmarked_phrase] = len(
                re.findall(unmarked_search, whole_text, flags=search_flags)
            )
            if unmarked_count[unmarked_phrase] > 0:
                wf_dialog.add_entry(
                    unmarked_phrase, unmarked_count[unmarked_phrase], suspect=True
                )
                suspect_cnt += 1

        if unmarked_count[unmarked_phrase] > 0 or not preferences.get(
            PrefKey.WFDIALOG_SUSPECTS_ONLY
        ):
            wf_dialog.add_entry(marked_phrase, marked_count)

    wf_dialog.display_entries()
    wf_dialog.message.set(
        f"{sing_plur(total_cnt, 'marked-up phrase')}; {sing_plur(suspect_cnt, 'suspect')} ({WordFrequencyEntry.SUSPECT})"
    )


def wf_populate_accents(wf_dialog: WordFrequencyDialog) -> None:
    """Populate the WF dialog with the list of all accented words.

    Args:
        wf_dialog: The word frequency dialog.
    """
    assert _the_word_lists is not None
    wf_dialog.reset()

    all_words = _the_word_lists.get_all_words()
    suspect_cnt = 0
    total_cnt = 0
    for word, freq in all_words.items():
        no_accent_word = DiacriticRemover.remove_diacritics(word)
        if no_accent_word != word:
            total_cnt += 1
            word_output = False
            if not preferences.get(PrefKey.WFDIALOG_SUSPECTS_ONLY):
                wf_dialog.add_entry(word, freq)
                word_output = True
            # Check for suspect, i.e. also seen without accents
            if no_accent_word in all_words:
                if not word_output:
                    wf_dialog.add_entry(word, freq)
                wf_dialog.add_entry(
                    no_accent_word, all_words[no_accent_word], suspect=True
                )
                suspect_cnt += 1
    wf_dialog.display_entries()
    wf_dialog.message.set(
        f"{sing_plur(total_cnt, 'accented word')}; {sing_plur(suspect_cnt, 'suspect')} ({WordFrequencyEntry.SUSPECT})"
    )


def wf_populate_ligatures(wf_dialog: WordFrequencyDialog) -> None:
    """Populate the WF dialog with the list of all ligature words.

    Args:
        wf_dialog: The word frequency dialog.
    """
    assert _the_word_lists is not None
    wf_dialog.reset()

    all_words = _the_word_lists.get_all_words()
    suspect_cnt = 0
    total_cnt = 0
    suspects_only = preferences.get(PrefKey.WFDIALOG_SUSPECTS_ONLY)
    for word, freq in all_words.items():
        if not re.search("(ae|AE|Ae|oe|OE|Oe|æ|Æ|œ|Œ)", word):
            continue
        total_cnt += 1
        # If actual ligature, only output it here if not suspects-only
        if re.search("(æ|Æ|œ|Œ)", word):
            if not suspects_only:
                wf_dialog.add_entry(word, freq)
        # Use the non-ligature version to check for suspects - because AE and Ae are both Æ
        else:
            lig_word = re.sub("ae", "æ", word)
            lig_word = re.sub("(AE|Ae)", "Æ", lig_word)
            lig_word = re.sub("oe", "œ", lig_word)
            lig_word = re.sub("(OE|Oe)", "Œ", lig_word)
            if lig_word in all_words:
                suspect_cnt += 1
                if suspects_only:
                    wf_dialog.add_entry(lig_word, all_words[lig_word])
                wf_dialog.add_entry(word, freq, suspect=True)
            elif not suspects_only:
                wf_dialog.add_entry(word, freq)
    wf_dialog.display_entries()
    wf_dialog.message.set(
        f"{sing_plur(total_cnt, 'ligature word')}; {sing_plur(suspect_cnt, 'suspect')} ({WordFrequencyEntry.SUSPECT})"
    )


def wf_populate_charcounts(wf_dialog: WordFrequencyDialog) -> None:
    """Populate the WF dialog with the list of all the characters.

    Args:
        wf_dialog: The word frequency dialog.
    """
    wf_dialog.reset()

    char_dict: WFDict = WFDict()

    total_cnt = 0
    for line, _ in maintext().get_lines():
        total_cnt += len(line)
        for char in line:
            tally_word(char_dict, char)

    for char, count in char_dict.items():
        wf_dialog.add_entry(char, count)

    wf_dialog.display_entries()
    wf_dialog.message.set(
        f"{sing_plur(total_cnt, 'character')}; {sing_plur(len(char_dict), 'distinct character')}"
    )


def wf_populate_regexps(wf_dialog: WordFrequencyDialog) -> None:
    """Populate the WF dialog with the list of all words that match the regexp.

    Args:
        wf_dialog: The word frequency dialog.
    """
    wf_dialog.reset()

    regexp = wf_dialog.regex_box.get()
    wf_dialog.regex_box.add_to_history(regexp)
    try:
        wf_populate_by_match(
            wf_dialog,
            "regex-matching",
            lambda word: re.search(regexp, word),
        )
    except re.error as exc:
        wf_dialog.message.set("Bad regex: " + str(exc))


def tally_word(wf_dict: WFDict, word: str) -> None:
    """Tally word in given WF dictionary, unless word is empty.

    If word already in dictionary, increment the count.
    If not, add word and set count to 1.

    Args:
        wf_dict: WFDict to tally word in.
        word: Word to be tallied."""
    if word:
        try:
            wf_dict[word] += 1
        except KeyError:
            wf_dict[word] = 1
